""" OPML.py

"""

__copyright__ = "Copyright (c) 2002-2005 Free Software Foundation, Inc."
__author__  = "Juri Pakaste <juri@iki.fi>"
__license__ = """
Straw is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

Straw is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA. """


from xml.sax import saxutils, make_parser, SAXParseException
from xml.sax.handler import feature_namespaces, feature_namespace_prefixes
from xml.sax.saxutils import XMLGenerator
from xml.sax.xmlreader import AttributesImpl
import xml.sax._exceptions
import xml.sax.handler
import sys

class OPML(dict):
    def __init__(self):
        self.outlines = []
    
    def output(self, stream = sys.stdout):
        xg = XMLGenerator(stream, encoding='utf-8')
        def elemWithContent(name, content):
            xg.startElement(name, AttributesImpl({}))
            if content is not None:
                xg.characters(content)
            xg.endElement(name)
            xg.characters("\n")
        xg.startElement("opml", AttributesImpl({'version': '1.1'}))
        xg.startElement("head", AttributesImpl({}))
        for key in ('title', 'dateCreated', 'dateModified', 'ownerName',
                    'ownerEmail', 'expansionState', 'vertScrollState',
                    'windowTop', 'windowBotton', 'windowRight', 'windowLeft'):
            if self.has_key(key) and self[key] != "":
                elemWithContent(key, self[key])
        xg.endElement("head")
        xg.startElement("body", AttributesImpl({}))
        for o in self.outlines:
            o.output(xg)
        xg.endElement("body")
        xg.endElement("opml")

class Outline(dict):
    __slots__ = ('_children')
    
    def __init__(self):
        self._children = []

    def add_child(self, outline):
        self._children.append(outline)

    def get_children_iter(self):
        return self.OIterator(self)

    children = property(get_children_iter, None, None, "")

    def output(self, xg):
        xg.startElement("outline", AttributesImpl(self))
        for c in self.children:
            c.output(xg)
        xg.endElement("outline")
        xg.characters("\n")

    class OIterator:
        def __init__(self, o):
            self._o = o
            self._index = -1

        def __iter__(self):
            return self

        def next(self):
            self._index += 1
            if self._index < len(self._o._children):
                return self._o._children[self._index]
            else:
                raise StopIteration

class OutlineList:
    def __init__(self):
        self._roots = []
        self._stack = []
    
    def add_outline(self, outline):
        if len(self._stack):
            self._stack[-1].add_child(outline)
        else:
            self._roots.append(outline)
        self._stack.append(outline)

    def close_outline(self):
        if len(self._stack):
            del self._stack[-1]

    def roots(self):
        return self._roots

class OPMLHandler(xml.sax.handler.ContentHandler):
    def __init__(self):
        self._outlines = OutlineList()
        self._opml = None
        self._content = ""

    def startElement(self, name, attrs):
        if self._opml is None:
            if name != 'opml':
                raise ValueError, "This doesn't look like OPML"
            self._opml = OPML()
        if name == 'outline':
            o = Outline()
            o.update(attrs)
            self._outlines.add_outline(o)
        self._content = ""

    def endElement(self, name):
        if name == 'outline':
            self._outlines.close_outline()
            return
        if name == 'opml':
            self._opml.outlines = self._outlines.roots()
            return
        for key in ('title', 'dateCreated', 'dateModified', 'ownerName',
                    'ownerEmail', 'expansionState', 'vertScrollState',
                    'windowTop', 'windowBotton', 'windowRight', 'windowLeft'):
            if name == key:
                self._opml[key] = self._content
                return
        
    def characters(self, ch):
        self._content += ch

    def get_opml(self):
        return self._opml

def parse(stream):
    parser = make_parser()
    parser.setFeature(feature_namespaces, 0)
    handler = OPMLHandler()
    parser.setContentHandler(handler)

    parser.parse(stream)
    return handler.get_opml()




""" utils.py

Module for utility methods.
"""
__copyright__ = "Copyright (c) 2002-2005 Free Software Foundation, Inc."
__license__ = """GNU General Public License

This program is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA. """

import re, string, htmlentitydefs
from htmlentitydefs import codepoint2name
import urllib, urlparse, locale
import time, calendar
import os
import os.path
import subprocess
import sys
import error
import constants
import gtk

entity = re.compile(r'\&.\w*?\;')


def convert_entities(text):
    def conv(ents):
        entities = htmlentitydefs.entitydefs
        ents = ents.group(0)
        ent_code = entities.get(ents[1:-1], None)
        if ent_code is not None:
            try:
                ents = unicode(ent_code, get_locale_encoding())
            except UnicodeDecodeError:
                ents = unicode(ent_code, 'latin-1')
            except Exception, ex:
                error.log("error occurred while converting entity %s: %s" % (ents, ex))

            # check if it still needs conversion
            if entity.search(ents) is None:
                return ents

        if ents[1] == '#':
            code = ents[2:-1]
            base = 10
            if code[0] == 'x':
                code = code[1:]
                base = 16
            return unichr(int(code, base))
        else:
            return

    in_entity = entity.search(text)
    if in_entity is None:
        return text
    else:
        ctext = in_entity.re.sub(conv, text)
        return ctext

def complete_url(url, feed_location):
    url = urllib.quote(url, safe=string.punctuation)
    if urlparse.urlparse(url)[0] == '':
        return urlparse.urljoin(feed_location, url)
    else:
        return url

def get_url_location(url):
    url = urllib.quote(url, safe=string.punctuation)
    parsed_url = urlparse.urlsplit(url)
    return urlparse.urlunsplit((parsed_url[0], parsed_url[1], '','',''))

def get_locale_encoding():
    try:
        encoding = locale.getpreferredencoding()
    except locale.Error:
        encoding = sys.getdefaultencoding()
    return encoding

def format_date(date, format=None, encoding=None):
    if format is None:
        format = get_date_format()
    if encoding is None:
        encoding = get_locale_encoding()
    timestr = time.strftime(format, time.localtime(calendar.timegm(date)))
    return unicode(timestr, encoding)

def get_date_format():
    # this is here just to make xgettext happy: it should be defined in
    # only one place, and a good one would be MainWindow.py module level.
    # however, we can't access _ there.
    # The format: %A is the full weekday name
    #             %B is the full month name
    #             %e is the day of the month as a decimal number,
    #                without leading zero
    # This should be translated to be suitable for the locale in
    # question, feel free to alter the order and the parameters (they
    # are strftime(3) parameters, the whole string is passed to the
    # function, Straw does no additional interpretation) if you feel
    # it's necessary in the translation file.
    return _('%A %B %e %H:%M')

def find_data_dir():
    datad = ''
    if os.environ.has_key("STRAW_DATA"):
        datad = os.environ["STRAW_DATA"]
    else:
        datad = constants.datadir
        if not os.path.isdir(datad):
            raise "FileNotFoundError", "couldn't find Straw data directory"
    return datad

def find_locale_dir():
    localed = ''
    if os.environ.has_key("STRAW_LOCALE"):
        localed = os.environ["STRAW_LOCALE"]
    else:
        localed = constants.localedir
        if not os.path.isdir(localed):
            raise "FileNotFoundError", "couldn't find locale data directory"
    return localed

def find_glade_file(libdir=None):
    gladef = ''
    if os.environ.has_key("STRAW_GLADE"):
        gladef = os.environ["STRAW_GLADE"]
    else:
        gladef = os.path.normpath(os.path.join(constants.gladedir, "straw.glade"))
        if not os.path.isfile(gladef):
            raise "FileNotFoundError", "couldn't find straw.glade"
    return gladef

def find_image_dir():
    if not os.path.isdir(constants.imagedir):
        raise "FileNotFoundError", "could not find image (pixmaps) directory"
    return constants.imagedir

def listdiff(l1, l2, test=None):
    if test is not None:
        return _listdifftest(l1, l2, test)
    common = []
    inl1 = []
    inl2 = []
    for e in l1:
        if e in l2:
            common.append(e)
        else:
            inl1.append(e)
    for e in l2:
        if e in l1:
            if e not in common:
                common.append(e)
        else:
            inl2.append(e)
    return (common, inl1, inl2)

import string
import gnomevfs
import Config
import dialogs

def url_show(url):
    config = Config.get_instance()
    if config.browser_cmd:
        try:
            cmdbin, args = string.splitfields(str(config.browser_cmd), maxsplit=1)
            link = args % url
            pid = subprocess.Popen([cmdbin, link]).pid
            return pid
        except ValueError, ve:
            dialogs.report_error(_("An error occurred while trying to open link"),
                                 _("There was a problem opening '%s'\n\nError thrown is '%s'") % (url,str(ve)))
    else:
        return gnomevfs.url_show(url)


def set_clipboard_text(text):
    clipboard = gtk.clipboard_get(selection="CLIPBOARD")
    clipboard.set_text(text)

def html_replace(exc):
    """ Python Cookbook 2ed, Section 1.23
    """
    if isinstance(exc, (UnicodeDecodeError, UnicodeTranslateError)):
        s = [ u'&%s;' % codepoint2name[ord(c)] for c in exc.object[exc.start:exc.end]]
        return ''.join(s), exc.end
    else:
        raise TypeError("can't handle %s" % exc.__name__)

import codecs
codecs.register_error('html_replace', html_replace)



""" Config.py

Module for Straw's configuration settings.
"""
__copyright__ = "Copyright (c) 2002-2005 Free Software Foundation, Inc."
__license__ = """ GNU General Public License

This program is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA. """

import cPickle, os, traceback
from urlparse import urlparse
from error import log, logtb, logparam
import pygtk
pygtk.require('2.0')
import gconf
import Event
import LookupManager

GCONF_STRAW_ROOT = "/apps/straw"

OPTION_LAST_POLL = "/general/last_poll"
OPTION_ITEMS_STORED = "/general/number_of_items_stored"
OPTION_ITEM_ORDER = "/general/item_order_newest"
OPTION_BROWSER_CMD = "/general/browser_cmd"
OPTION_WINDOW_SIZE_W = "/ui/window_width"
OPTION_WINDOW_SIZE_H = "/ui/window_height"
OPTION_MAIN_PANE_POS = "/ui/main_pane_position"
OPTION_SUB_PANE_POS = "/ui/sub_pane_position"
OPTION_WINDOW_MAX = "/ui/window_maximized"
OPTION_MAGNIFICATION = "/ui/text_magnification"
OPTION_OFFLINE = "/general/offline"
OPTION_POLL_FREQUENCY = "/general/poll_frequency"
OPTION_FEED_ID_SEQ = "feed_id_seq"
OPTION_FEEDS = "feeds"
OPTION_CATEGORIES = "categories"

def ensure_directory(strawdir):
    if os.path.exists(strawdir):
        if not os.path.isdir(strawdir):
            raise Exception
        return 0
    os.mkdir(strawdir)
    return 1

#############
# Config persistence classes
class ConfigPicklePersistence:
    def __init__(self, filename):
        self._config_file = filename
        self._dict = None

    def save_option(self, option, value):
        if self._dict is None:
            self._initialize_dict()
        temp_config_file = self._config_file + ".working"
        pickle_file = open(temp_config_file, "w")
        self._dict[option] = value
        cPickle.dump(self._dict, pickle_file, True)
        pickle_file.close()
        os.rename(temp_config_file, self._config_file)
        return

    def load_option(self, option):
        if self._dict is None:
            self._initialize_dict()
        return self._dict.get(option, None)

    def _initialize_dict(self):
        if os.path.exists(self._config_file):
            pickle_file = open(self._config_file, "r")
            self._dict = cPickle.load(pickle_file)
            pickle_file.close()
        else:
            self._dict = {}

class ConfigGConfPersistence:
    SAVERS = {OPTION_LAST_POLL: 'int',
              OPTION_ITEMS_STORED: 'int',
              OPTION_ITEM_ORDER: 'bool',
              OPTION_BROWSER_CMD: 'string',
              OPTION_WINDOW_SIZE_W: 'int',
              OPTION_WINDOW_SIZE_H: 'int',
              OPTION_MAIN_PANE_POS: 'int',
              OPTION_SUB_PANE_POS: 'int',
              OPTION_OFFLINE: 'bool',
              OPTION_WINDOW_MAX: 'bool',
              OPTION_MAGNIFICATION: 'float',
              OPTION_POLL_FREQUENCY: 'int'}

    def __init__(self, client):
        self.client = client

    def save_option(self, option, value):
        getattr(self.client, 'set_' + self.SAVERS[option])(
            GCONF_STRAW_ROOT + option, value)

    def load_option(self, option):
        return getattr(self.client, 'get_' + self.SAVERS[option])(
            GCONF_STRAW_ROOT + option)

class ConfigPersistence:
    def __init__(self, *backends):
        self.backends = backends

    def save_option(self, option, value):
        for b in self.backends:
            if option in b[1]:
                b[0].save_option(option, value)

    def load_option(self, option):
        for b in self.backends:
            if option in b[1]:
                return b[0].load_option(option)

################
# Proxy config
class BaseProxyConfig(object):
    def __init__(self):
        self._config_working = False

    @property
    def config_working(self):
        return self._config_working

    @apply
    def host():
        doc = ""
        def fget(self):
            return self._host
        def fset(self, host):
            self._host = host
            self._ip = None
        return property(**locals())

    @apply
    def ip():
        doc = ""
        def fget(self):
            return self._ip
        def fset(self, ip):
            assert self._ip is None, "Overwriting proxy IP not allowed!"
            if ip is None:
                self.use = 0
            self._ip = ip
        return property(**locals())

    @property
    def is_waiting(self):
        return self.use and not self._ip

    def _ip_lookup_callback(self, host, ip, data):
        self.ip = ip

    def lookup_host(self):
        if self._host and not self._ip:
            LookupManager.get_instance().lookup(
                self._host, self._ip_lookup_callback)


class GconfProxyConfig(BaseProxyConfig):
    """Encapsulate proxy use and location information (host, port, ip),
    gconf reading and name lookup logic"""

    GCONF_PROXY_ROOT = "/system/proxy"
    GCONF_PROXY_MODE = GCONF_PROXY_ROOT + "/mode"
    GCONF_HTTP_PROXY_ROOT = "/system/http_proxy"
    GCONF_HTTP_PROXY_HOST = GCONF_HTTP_PROXY_ROOT + "/host"
    GCONF_HTTP_PROXY_PORT = GCONF_HTTP_PROXY_ROOT + "/port"
    GCONF_HTTP_PROXY_USE_AUTHENTICATION = GCONF_HTTP_PROXY_ROOT + "/use_authentication"
    GCONF_HTTP_PROXY_USER = GCONF_HTTP_PROXY_ROOT + "/authentication_user"
    GCONF_HTTP_PROXY_PASSWORD = GCONF_HTTP_PROXY_ROOT + "/authentication_password"

    def __init__(self):
        BaseProxyConfig.__init__(self)
        client = gconf.client_get_default()
        self._config_working = client.dir_exists(self.GCONF_PROXY_ROOT) and client.dir_exists(self.GCONF_HTTP_PROXY_ROOT)
        if not self._config_working:
            return
        client.add_dir(self.GCONF_PROXY_ROOT, gconf.CLIENT_PRELOAD_RECURSIVE)
        client.add_dir(self.GCONF_HTTP_PROXY_ROOT, gconf.CLIENT_PRELOAD_RECURSIVE)
        client.notify_add(self.GCONF_PROXY_MODE, self.proxy_mode_changed)
        client.notify_add(self.GCONF_HTTP_PROXY_HOST, self.proxy_host_changed)
        client.notify_add(self.GCONF_HTTP_PROXY_PORT, self.proxy_port_changed)
        client.notify_add(self.GCONF_HTTP_PROXY_USE_AUTHENTICATION,
                          self.proxy_auth_changed)
        client.notify_add(self.GCONF_HTTP_PROXY_USER, self.proxy_auth_changed)
        client.notify_add(self.GCONF_HTTP_PROXY_PASSWORD, self.proxy_auth_changed)

        self._ip = None
        pm = client.get_string(self.GCONF_PROXY_MODE)
        self.use = pm is not None and pm != "none"
        self.host = client.get_string(self.GCONF_HTTP_PROXY_HOST)
        self.port = client.get_int(self.GCONF_HTTP_PROXY_PORT)
        self.use_authentication = client.get_bool(
            self.GCONF_HTTP_PROXY_USE_AUTHENTICATION)
        self.user = client.get_string(self.GCONF_HTTP_PROXY_USER)
        self.password = client.get_string(self.GCONF_HTTP_PROXY_PASSWORD)

    # here be gconf logic
    def proxy_mode_changed(self, client, notify_id, entry, *args):
        value = entry.value.get_string()
        self.use = value is not None and value != "none"

    def proxy_host_changed(self, client, notify_id, entry, *args):
        value = entry.value.get_string()
        if value is not None and len(value):
            self.host = value
        else:
            self.use = 0

    def proxy_port_changed(self, client, notify_id, entry, *args):
        value = entry.value.get_int()
        if value is not None and value > 0:
            self.port = value
        else:
            self.use = 0

    def proxy_auth_changed(self, client, notify_id, entry, *args):
        value = entry.value
        if entry.key == self.GCONF_HTTP_PROXY_USE_AUTHENTICATION:
            if not value:
                self.use_authentication = 0
            else:
                self.use_authentication = 1
        elif entry.key == self.GCONF_HTTP_PROXY_USER:
            self.user = value
        elif entry.key == self.GCONF_HTTP_PROXY_PASSWORD:
            self.password = value

class EnvironmentProxyConfig(BaseProxyConfig):
    """Encapsulate proxy use and location information, environment reading"""

    def __init__(self):
        BaseProxyConfig.__init__(self)
        try:
            result = self._read_env()
        except:
            result = 0
        self._config_working = result
        self.use = self._config_working

    def _read_env(self):
        proxy = os.getenv("http_proxy")
        if not proxy:
            return 0
        pparts = urlparse(proxy)
        proto = pparts[0]
        host = pparts[1]
        if proto != "http" or not len(host):
            return 0
        hparts = host.split('@')
        auth = self.user = self.password = self.use_authentication = None
        if len(hparts) > 2:
            return 0
        if len(hparts) == 2:
            auth, host = hparts
        if auth:
            aparts = auth.split(':')
            if len(aparts) != 2:
                return 0
            self.user, self.password = aparts
            self.use_authentication = 1
        self.host = host
        self.port = 80
        hparts = host.split(':')
        if len(hparts) > 2:
            return 0
        if len(hparts) == 2:
            self.host = hparts[0]
            self.port = int(hparts[1])
        return 1

class NullProxyConfig(BaseProxyConfig):
    def __init__(self):
        BaseProxyConfig.__init__(self)
        self.use = 0
        self._config_working = 0
        self._host = None
        self._ip = None
        self.port = None
        self.user = None
        self.password = None
        self.use_authentication = None

###################
# The main man
class Config(object, Event.SignalEmitter):
    _straw_dir = os.path.join(os.getenv("HOME"), ".straw")
    _straw_config_file = os.path.join(_straw_dir, "config")

    def __init__(self, persistence):
        Event.SignalEmitter.__init__(self)
        self.persistence = persistence
        self.initialize_slots(Event.ItemOrderChangedSignal,
                              Event.OfflineModeChangedSignal,
                              Event.PollFrequencyChangedSignal,
                              Event.NumberOfItemsStoredChangedSignal)
        self._feed_id_seq = 0
        self._poll_freq = 1800
        self._last_poll = 0
        self._browser_cmd = ''
        self._items_stored = 30
        self._item_order = False
        self._main_window_size = (640,480)
        self._main_pane_position = 100
        self._sub_pane_position = 100
        self.first_time = None
        self._offline = True
        self._window_maximized = False
        self._use_threads = True
        self._reload_css = False
        self._no_etags = False

    def initialize_proxy(self):
        pcchoices = (GconfProxyConfig, EnvironmentProxyConfig, NullProxyConfig)
        for p in pcchoices:
            self._proxy_config = p()
            if self._proxy_config.config_working:
                break

    def initialize_options(self):
        # Call this after calling Config's constructor
        self.first_time = ensure_directory(self._straw_dir)

        if os.path.exists(self.straw_config_file):
            self._feed_id_seq = self.persistence.load_option(OPTION_FEED_ID_SEQ)

        self._poll_freq = self.persistence.load_option(OPTION_POLL_FREQUENCY)
        self._last_poll = self.persistence.load_option(OPTION_LAST_POLL)
        self._items_stored = self.persistence.load_option(OPTION_ITEMS_STORED)
        self._browser_cmd = self.persistence.load_option(OPTION_BROWSER_CMD)
        self._item_order = self.persistence.load_option(OPTION_ITEM_ORDER)

        width = self.persistence.load_option(OPTION_WINDOW_SIZE_W)
        height = self.persistence.load_option(OPTION_WINDOW_SIZE_H)
        if width <= 0:
            width = 640
        if height <= 0:
            height = 480
        self._main_window_size = (width, height)

        self._main_pane_position = self.persistence.load_option(
            OPTION_MAIN_PANE_POS)
        self._sub_pane_position = self.persistence.load_option(
            OPTION_SUB_PANE_POS)
        self._window_maximized = self.persistence.load_option(OPTION_WINDOW_MAX)
        self._text_magnification = self.persistence.load_option(OPTION_MAGNIFICATION)
        self._offline = self.persistence.load_option(OPTION_OFFLINE)

    @property
    def straw_dir(self):
        return self._straw_dir

    @property
    def straw_config_file(self):
        return self._straw_config_file

    @property
    def proxy_config(self):
        return self._proxy_config

    @apply
    def feeds():
        doc = "Marshalled feed data"
        def fget(self):
            return self.persistence.load_option(OPTION_FEEDS)
        def fset(self, feeddata):
            self.persistence.save_option(
                OPTION_FEEDS, feeddata)
        return property(**locals())

    @apply
    def categories():
        doc = ""
        def fget(self):
            return self.persistence.load_option(OPTION_CATEGORIES)
        def fset(self, categorydata):
            self.persistence.save_option(
                OPTION_CATEGORIES, categorydata)
        return property(**locals())
    
    @apply
    def poll_frequency():
        doc = "Polling frequency"
        def fget(self):
            return self._poll_freq
        def fset(self, poll_frequency):
            if self._poll_freq != poll_frequency:
                self._poll_freq = poll_frequency
                self.persistence.save_option(OPTION_POLL_FREQUENCY, poll_frequency)
                self.emit_signal(Event.PollFrequencyChangedSignal(self, poll_frequency))
        return property(**locals())

    @apply
    def last_poll():
        doc = "last poll"
        def fget(self):
            return self._last_poll
        def fset(self, last_poll):
            if self._last_poll != last_poll:
                self._last_poll = last_poll
                self.persistence.save_option(OPTION_LAST_POLL, last_poll)
        return property(**locals())

    @apply
    def  browser_cmd():
        doc = "The browser to use"
        def fget(self):
            return self._browser_cmd
        def fset(self, browser_cmd):
            if self._browser_cmd != browser_cmd:
                self._browser_cmd = browser_cmd
                self.persistence.save_option(OPTION_BROWSER_CMD, browser_cmd)
        return property(**locals())

    @apply
    def number_of_items_stored():
        doc = "Number of items to store per feed"
        def fget(self):
            return self._items_stored
        def fset(self, num=30):
            if self._items_stored != num:
                self._items_stored = num
                self.persistence.save_option(OPTION_ITEMS_STORED, num)
                self.emit_signal(Event.NumberOfItemsStoredChangedSignal(self))
        return property(**locals())

    @apply
    def item_order():
        doc = "Ordering of Items"
        def fget(self):
            return self._item_order
        def fset(self, order):
            if self._item_order != order:
                self._item_order = order
                self.persistence.save_option(OPTION_ITEM_ORDER, order)
                self.emit_signal(Event.ItemOrderChangedSignal(self))
        return property(**locals())

    @apply
    def feed_id_seq():
        doc = ""
        def fget(self):
            return self._feed_id_seq
        def fset(self, id):
            self._feed_id_seq = id
            self.persistence.save_option(OPTION_FEED_ID_SEQ, id)
        return property(**locals())

    def next_feed_id_seq(self):
        self.feed_id_seq += 1
        return self._feed_id_seq

    @apply
    def main_window_size():
        doc = ""
        def fget(self):
            return self._main_window_size
        def fset(self, size):
            if self._main_window_size != size:
                self._main_window_size = size
                self.persistence.save_option(OPTION_WINDOW_SIZE_W, size[0])
                self.persistence.save_option(OPTION_WINDOW_SIZE_H, size[1])
        return property(**locals())

    @apply
    def offline():
        doc = ""
        def fget(self):
            return self._offline
        def fset(self, mode):
            if self._offline != mode:
                self._offline = mode
                self.persistence.save_option(OPTION_OFFLINE, mode)
                self.emit_signal(Event.OfflineModeChangedSignal(self))
        return property(**locals())

    @apply
    def window_maximized():
        doc = ""
        def fget(self):
            return self._window_maximized
        def fset(self, state):
            if self._window_maximized is not state:
                self._window_maximized = state
                self.persistence.save_option(OPTION_WINDOW_MAX, state)
        return property(**locals())

    @apply
    def main_pane_position():
        doc = ""
        def fget(self):
            return self._main_pane_position
        def fset(self, position):
            if self._main_pane_position != position:
                self._main_pane_position = position
                self.persistence.save_option(OPTION_MAIN_PANE_POS, position)
        return property(**locals())

    @apply
    def sub_pane_position():
        doc = ""
        def fget(self):
            return self._sub_pane_position
        def fset(self, position):
            if self._sub_pane_position != position:
                self._sub_pane_position = position
                self.persistence.save_option(OPTION_SUB_PANE_POS, position)
        return property(**locals())

    @apply
    def text_magnification():
        doc = "sets the amount of magnification in the item view"
        def fget(self):
            return self._text_magnification
        def fset(self, tm):
            if self._text_magnification is not tm:
                self._text_magnification = tm
                self.persistence.save_option(OPTION_MAGNIFICATION, tm)
        return property(**locals())

    @apply
    def use_threads():
        doc = ""
        def fget(self):
            return self._use_threads
        def fset(self, threads):
            self._use_threads = threads
        return property(**locals())

    @apply
    def reload_css():
        doc = ""
        def fget(self):
            return self._reload_css
        def fset(self, reload_css):
            self._reload_css = reload_css
        return property(**locals())

    @apply
    def no_etags():
        doc = ""
        def fget(self):
            return self._no_etags
        def fset(self, no_etags):
            self._no_etags = no_etags
        return property(**locals())

def convert_if_necessary(config):
    if not os.path.exists(config.straw_config_file):
        return

    try:
        f = open(config.straw_config_file, "r")
        cf = cPickle.load(f)

        feeds = cf.get('feeds', None)
        if feeds and feeds[0].get('_n_items_unread', None) is None:

            def _convert_feeds(feeds, parent):
                import ItemStore
                itemstore = ItemStore.get_instance(config.straw_dir)
                for feed in feeds:
                    if isinstance(feed, list):
                        _convert_feeds(feed[1:], parent)
                    else:
                        stored = feed.get('_items_stored', -1)
                        if stored > -1:
                            cutoff = stored
                        else:
                            cutoff = config.number_of_items_stored
                        feed['_n_items_unread'] = itemstore.get_number_of_unread(feed['_id'], cutoff)
                return feeds

            config.feeds = _convert_feeds(feeds, None)

        if cf.has_key('poll_frequency'):
            config.poll_frequency = cf.get('poll_frequency')
            config.number_of_items_stored = cf.get('number_of_items_stored')
            config.item_order = cf.get('item_order')
            config.main_window_size = cf.get('main_window_size')
            config.main_pane_position = cf.get('main_pane_position')
            config.sub_pane_position = cf.get('sub_pane_position')
            config.offline = cf.get('offline')

            del cf['poll_frequency']
            del cf['number_of_items_stored']
            del cf['item_order']
            del cf['main_window_size']
            del cf['main_pane_position']
            del cf['sub_pane_position']
            del cf['offline']

            f.close()
            f = open(config.straw_config_file, "w")
            f.seek(0)
            f.truncate()
            cPickle.dump(cf, f, True)
    finally:
        f.close()
    return

def create_gconf_persistence():
    client = gconf.client_get_default()
    client.add_dir(GCONF_STRAW_ROOT, gconf.CLIENT_PRELOAD_ONELEVEL)
    return ConfigGConfPersistence(client)

def create_pickle_persistence():
    return ConfigPicklePersistence(Config._straw_config_file)

def create_instance():
    # Add your GConf key here as well!
    cp = ConfigPersistence(
        (create_gconf_persistence(),
         (OPTION_LAST_POLL, OPTION_ITEMS_STORED, OPTION_ITEM_ORDER, OPTION_BROWSER_CMD,
          OPTION_WINDOW_SIZE_W, OPTION_WINDOW_SIZE_H,
          OPTION_MAIN_PANE_POS, OPTION_SUB_PANE_POS, OPTION_OFFLINE,
          OPTION_MAGNIFICATION,
          OPTION_POLL_FREQUENCY, OPTION_WINDOW_MAX)),
        (create_pickle_persistence(),
         (OPTION_FEED_ID_SEQ, OPTION_FEEDS, OPTION_CATEGORIES)))
    config = Config(cp)
    config.initialize_proxy()
    config.initialize_options()
    convert_if_necessary(config)
    return config

config_instance = None
def get_instance():
    global config_instance
    if config_instance is None:
        config_instance = create_instance()
    return config_instance



""" OfflineToggle.py

"""
__copyright__ = "Copyright (c) 2002-2005 Free Software Foundation, Inc."
__license__ = """
Straw is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

Straw is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA. """


import pygtk
pygtk.require('2.0')
import gtk
import os
import MVP
import Event
import utils
import Config

class OfflineView(MVP.WidgetView):
    """
    Model: Config instance
    Presenter: OfflinePresenter instance
    """
    def _initialize(self):
        self._tooltips = gtk.Tooltips()
        self._tooltips.enable()

    def _on_offline_toggle_toggled(self, *args):
        status = self._widget.get_active()
        if status != self._model.offline:
            self._model.offline = status

    def set_status(self, active):
        self._widget.set_active(active)

    def set_text(self, text):
        self._tooltips.set_tip(self._widget, text, text)

    def set_icon(self, icon):
        image = self._widget.get_child()
        image.set_from_stock(icon, gtk.ICON_SIZE_BUTTON)

class OfflinePresenter(MVP.BasicPresenter):
    """
    Model: Config instance
    View: OfflineView instance
    """

    def _initialize(self):
        self._attribs = {'online': (gtk.STOCK_CONNECT, _("Straw is currently online. Click to go offline.")),
                         'offline': (gtk.STOCK_DISCONNECT, _("Straw is currently offline. Click to go online."))}
        self._model.signal_connect(Event.OfflineModeChangedSignal,
                                   self._config_offline_changed)
        if self._model.offline:
            self._config_offline_changed(None)

    def _config_offline_changed(self, signal):
        offline = self._model.offline
        if offline:
            icon, tip = self._attribs.get('offline')
        else:
            icon, tip = self._attribs.get('online')
        self._view.set_status(offline)
        self._view.set_text(tip)
        self._view.set_icon(icon)

class OfflineToggle:
    def __init__(self, widget):
        config = Config.get_instance()
        oview =  OfflineView(widget, model=config)
        self._presenter = OfflinePresenter(model=config, view=oview)




import unittest
import os
import sys

try:
    import pmock
except ImportError:
    sys.exit("You need python-mock (pmock, 0.3) to run this test. http://pmock.sf.net")

sys.path.insert(0, '../src/lib')
import subscribe

class FeedLocationPresenterTestCase(unittest.TestCase):
    def setUp(self):
        self.presenter = subscribe.FeedLocationPresenter()

    def testSplit_urlWithAbsoluteUrl(self):
        feed = "http://www.unpluggable.com/plans/log.rdf"
        url, username, password = self.presenter.split_url(feed)
        self.assertEqual(url, feed)

    def testSplit_urlWithAuthToken(self):
        feed = "http://foo:com@test.org/very/long/path/to/feed.xml"
        url, username, password = self.presenter.split_url(feed)
        self.assertEqual(url, "http://test.org/very/long/path/to/feed.xml")
        self.assertEqual("foo", username)
        self.assertEqual("com", password)

    def testUrlWithLeadingSpace(self):
        feed = "  http://www.test.org/rss10.xml "
        url, username, password = self.presenter.split_url(feed)
        self.assertEqual("http://www.test.org/rss10.xml", url)

    def testUrlSchemeNotHttp(self):
        feed = "file:///home/bart/simpson.xml"
        view = pmock.Mock()
        model = pmock.Mock()
        view.expects(pmock.at_least_once()).method("report_error")
        model.expects(pmock.at_least_once()).method("poll")
        self.presenter.view = view
        self.presenter.model = model
        url, uname, pword = self.presenter.split_url(feed)
        self.presenter.find_feed(url)
        try:
            view.verify()
        except Exception, ex:
            # view.report_error() should've been called since
            # 'file' is not currently supported
            self.fail(ex)

    def testDomainCheck(self):
        feed = "http://www.foo.org/blog/devel/atom.xml"
        self.presenter.split_url(feed)
        self.assertEqual("www.foo.org", self.presenter.get_feed_domain())

    def tearDown(self):
        self.presenter = None

if __name__ == '__main__':
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(FeedLocationPresenterTestCase, 'test'))
    #suite.addTest(unittest.makeSuite(FeedsPresenterTestCase, 'test'))
    unittest.TextTestRunner(verbosity=2).run(suite)



""" Summaryparser.py

Wrapper module to feedparser and responsible for assigning data to Feed and
SummaryItems.
"""
__copyright__ = "Copyright (c) 2002-2005 Free Software Foundation, Inc."
__license__ = """
Straw is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

Straw is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA. """


import sys
import types
import copy
import htmlentitydefs
import HTMLParser
import string
import time

import feedparser
import error

import utils
import SummaryItem
import ParsedSummary

class TitleImgParser(HTMLParser.HTMLParser):
    def __init__(self, feed=None):
        HTMLParser.HTMLParser.__init__(self)
        self._chars = list()
        self._image_urls = list()
        self._feed = feed

    def set_feed(self, feed):
        self._feed = feed

    def get_image_urls(self):
        return self._image_urls

    def get_text(self, nchars=None):
        text = ''.join(self._chars).strip()
        if nchars:
            text = text[:nchars]
        return text

    def close(self):
        self.flush()
        HTMLParser.HTMLParser.close(self)

    def flush(self):
        del self._chars[:]
        del self._image_urls[:]

    def handle_starttag(self, tag, attrs):
        if tag == 'img':
            for name, value in attrs:
                if name == 'src':
                    url = utils.complete_url(value, self._feed.location)
                    self._image_urls.append(url)
        return

    def handle_data(self, data):
        self._chars.append(data)

    def handle_charref(self, ref):
        # called for each character reference, e.g. for '&#160;', ref will be '160'
        if not self._chars: return
        ref = ref.lower()
        if ref in ('34', '38', '39', '60', '62', 'x22', 'x26', 'x27', 'x3c', 'x3e'):
            text = '&#%s;' % ref
        else:
            if ref[0] == 'x':
                c = int(ref[1:], 16)
            else:
                c = int(ref)
            text = unichr(c).encode('utf-8')
        self._chars.append(text)

    def handle_entityref(self, ref):
        # called for each entity reference, e.g. for '&copy;', ref will be 'copy'
        if not self._chars: return
        if ref in ('lt', 'gt', 'quot', 'amp', 'apos'):
            text = '&%s;' % ref
        else:
            # entity resolution graciously donated by Aaron Swartz
            def name2cp(k):
                import htmlentitydefs
                if hasattr(htmlentitydefs, 'name2codepoint'): # requires Python 2.3
                    return htmlentitydefs.name2codepoint[k]
                k = htmlentitydefs.entitydefs[k]
                if k.startswith('&#') and k.endswith(';'):
                    return int(k[2:-1]) # not in latin-1
                return ord(k)
            try: name2cp(ref)
            except KeyError: text = '&%s;' % ref
            else: text = unichr(name2cp(ref)).encode('utf-8')
        self._chars.append(text)



def _remove_ids_if_duplicates(items):
    ids = {}
    duplicates = False
    for i in items:
        if i.guid is not None and i.guid != "":
            if ids.has_key(i.guid):
                duplicates = True
                break
            ids[i.guid] = True
    if duplicates:
        for i in items:
            i.guid = None
            i.guidislink = False
    return

def _to_unicode(text, encoding):
    if text and not isinstance(text, types.UnicodeType):
        text = unicode(text, encoding)
    return text

def feedparser_parse(data):
    pc = feedparser.parse(data)
    enc = pc.get('encoding', utils.get_locale_encoding())
    if not enc:
        enc = sys.getdefaultencoding()
    return (pc, enc)

def parse_channel_info(parsed, parsed_content, encoding):
    parsed.title = _to_unicode(parsed_content.feed.get('title', ''), encoding)
    parsed.description = _to_unicode(parsed_content.feed.get('description',''), encoding)
    parsed.link = _to_unicode(parsed_content.feed.get('link', ''), encoding)
    parsed.copyright = _to_unicode(parsed_content.feed.get('copyright',''), encoding)
    parsed.last_build_date = parsed_content.feed.get('modified')
    parsed.creator = _to_unicode(parsed_content.feed.get('creator', ''), encoding)
    return parsed

def parse(content, feed):
    parsed = ParsedSummary.ParsedSummary()
    parsed_content, encoding = feedparser_parse(content)
    parsed = parse_channel_info(parsed, parsed_content, encoding)
    for entry in parsed_content.entries:
        item = _parse_entry(entry, feed)
        parsed.addItem(item)
    _remove_ids_if_duplicates(parsed.items)
    return parsed

def sanitize_content(data, feed, limit=60):
    images = None
    title = ""
    try:
        tp = TitleImgParser(feed)
        try:
            tp.feed(data)
            images = [image for image in tp.get_image_urls()]
            title = tp.get_text(limit)
        except Exception, ex:
            error.log(ex)
    finally:
        tp.close()
    return (title, images)

def _parse_entry(entry, feed):
    item = SummaryItem.SummaryItem()
    item.feed = feed
    content = []
    description = ""
    title = _("No title")

    if entry.has_key('content'):
        # it can have multiple content, so we just aggregate them for now.
        for c in entry.content:
            try:
                if c.value not in content:
                    content.append(c.value)
            except TypeError, te:
                error.log(te)
                pass

    if not len(content) and entry.has_key('summary'):
        content.append(entry.get('summary',''))

    description = "<br/>".join(content)

    title = entry.get('title', '')
    if description:
        alttitle, images = sanitize_content(description, feed)
        [item.add_image(image) for image in images]
        if not title:
            # get the first MAXSPLIT words of the description and make that as our
            # title
            dwords = string.splitfields(alttitle, maxsplit=6)
            title = ' '.join(dwords[:]) + ' ...'
    title = title.replace('\n','')
    item.title = title

    item.description = description
    item.guidislink = entry.get('guidislink', False)
    item.link = entry.get('link', None)
    item.guid = entry.get('guid', None)
    item.creator = entry.get('author', None)
    item.contributors = entry.get('contributors', None)
    item.pub_date = entry.get('modified_parsed', time.localtime())
    item.license_urls.append(entry.get('license', None))
    item.fm_license = entry.get('fm_license', None)
    item.fm_changes = entry.get('fm_changes', None)
    item.publication_name = entry.get('prism_publicationname', None)
    item.publication_volume = entry.get('prism_volume', None)
    item.publication_number = entry.get('prism_number', None)
    item.publication_section = entry.get('prism_section',None)
    item.publication_starting_page = entry.get('prism_startingpage', None)
    item.enclosures = entry.get('enclosures', None)

    if entry.has_key('source'):
        url = entry.source.get('url', None)
        text = entry.source.get('value', None)
        if url and text:
            item.source = {'url': url,
                           'text': text}
        else:
            # There's no point displaying the source if there's no url in the
            # first place. This is a violation of the RSS 0.92 spec
            # http://backend.userland.com/rss092.
            item.source = None

    return item



""" httplib_async.py

Asynchronous HTTP client.
"""
__copyright__ = "Copyright (c) 2002-2005 Free Software Foundation, Inc."
__license__ = """ GNU General Public License

This program is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA. """


import asyncore
import socket
import mimetools
try:
    from cStringIO import StringIO
except:
    from StringIO import StringIO
import sys
from error import log

class HTTPConnection_async(asyncore.dispatcher_with_send):
    # asynchronous http client
    def __init__(self, host, ip, port, consumer, sock=None, map=None, proxy=None):
        asyncore.dispatcher_with_send.__init__(self)

        if ip is None:
            ip = host
        self.ip = ip
        self.host = host
        self.port = port
        self.consumer = consumer
        self.status = None
        self.http_version = "HTTP/1.1"
        self.proxy = None
        self.proxy_port = None
        if proxy:
            self.proxy = proxy[0]
            self.proxy_port = proxy[1]

    def set_proxy(self, proxy):
        self.proxy = proxy[0]
        self.proxy_port = proxy[1]

    def request(self, method, path, headers={}, data=None):
        self.header = None
        self.data = ""
        send_data = []
        self.method = method or "GET"
        self.path = path or "/"
        if self.path:
            if self.path[0] != "/":
                self.path = "/" + self.path
        else:
            self.path = "/"
        if self.proxy and self.proxy_port:
            self.path = "http://%s%s" % (self.host, path)
        self._initialize_chunked()
        send_data.append("%s %s %s\r\n" % (self.method, self.path, self.http_version))
        host_found = 0
        cl_found = 0
        for h, v in headers.items():
            h = h.lower()
            if h == 'host':
                host_found = 1
            if h == 'content-length':
                cl_found = 1
            send_data.append("%s: %s\r\n" % (h, v))
        if not host_found:
            send_data.append("Host: %s\r\n" % self.host)
        if data and not cl_found:
            send_data.append("Content-length: %d\r\n" % len(data))
        send_data.append("\r\n")
        if data:
            send_data.append(data)
        self.send_data = "".join(send_data)
        if self.connected:
            self.send(self.send_data)

    def execute(self):
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        if self.proxy and self.proxy_port:
            self.connect((self.proxy, self.proxy_port))
        else:
            self.connect((self.ip, self.port))

    def handle_connect(self):
        # connection succeeded
        self.send(self.send_data)

    def handle_expt(self):
        # This is used for OOB data.
        log("OOB data received")

    def handle_error(self):
        exctype, value = sys.exc_info()[:2]
        self.close()
        self.consumer.http_failed(value)

    def _initialize_chunked(self, chunked=0):
        self.chunked = chunked
        self.chunk_left = None
        self.chunk_data = []
        self.chunks_done = 0

    def handle_read(self):
        data = self.recv(4096)
        self.data += data

        if not self.header:
            # check if we've seen a full header

            header = self.data.split("\r\n\r\n", 1)
            if len(header) <= 1:
                # i guess not everyone can read the spec - congrats go to
                # thinkgeek
                header = self.data.split("\n\n", 1)
                if len(header) <= 1:
                    return
            header, data = header

            # parse header
            fp = StringIO(header)
            line = fp.readline()
            try:
                self.status = line.split(" ", 2)
                self.status[1] = int(self.status[1])
            except ValueError:
                self.consumer.http_failed(BadStatusException(line))
            self.header = mimetools.Message(fp)

            self.data = data
            if self.status[1] in (301, 302):
                self.consumer.http_redirect(self.header.getheader('location'),self.status[1] == 301)
                self.close()
                return
            elif self.header.getheader('transfer-encoding', '').lower() == 'chunked':
                self._initialize_chunked(1)
            else:
                self.consumer.http_header(self.status, self.header)

                if not self.connected:
                    return # channel was closed by consumer

        if self.chunked:
            # we are reading chunked data, read it all and the possible
            # extra headers after that
            while not self.chunks_done and len(self.data) > 0:
                if self.chunk_left is None or self.chunk_left < 0:
                    # one chunk over, read the length of the next one
                    if self.data[:2] == '\r\n':
                        self.data = self.data[2:]
                    cdata = self.data.split("\r\n", 1)
                    if len(cdata) <= 1:
                        return
                    line, self.data = cdata
                    i = line.find(';')
                    if i >= 0:
                        line = line[:i]
                    self.chunk_left = int(line, 16)
                    if self.chunk_left == 0:
                        # no more chunks, put back CRLF to help things later
                        self.data = "\r\n" + self.data
                        self.chunks_done = 1
                        break
                self.chunk_data.append(self.data[:self.chunk_left])
                l = len(self.data)
                self.data = self.data[self.chunk_left:]
                self.chunk_left -= l
            if self.chunks_done:
                cdata = self.data.split("\r\n\r\n", 1)
                if len(cdata) <= 1:
                    return
                headerlines = cdata[0][2:]
                fp = StringIO(headerlines)
                header = mimetools.Message(fp)
                chunk_data = "".join(self.chunk_data)
                for h, v in header.items():
                    self.header[h] = v
                if not self.header.has_key('content-length'):
                    self.header['content-length'] = str(len(chunk_data))
                del self.header['transfer-encoding']
                self.consumer.http_header(self.status, self.header)
                self.consumer.feed(chunk_data)
                return
        elif data:
            self.consumer.feed(data)

    def handle_close(self):
        self.close()
        self.consumer.http_close()
        self.consumer = None

class HTTPException(Exception):
    pass

class BadStatusException(HTTPException):
    def __init__(self, line):
        HTTPException.__init__(self)
        self.args = line
        self.line = line



""" OPMLExport.py
"""

__copyright__ = "Copyright (c) 2002-2005 Free Software Foundation, Inc."
__license__ = """
Straw is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

Straw is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA. """


import pygtk
pygtk.require('2.0')
import gnomevfs
import OPML

def export(title, list, fname):
    opml = OPML.OPML()
    opml['title'] = title
    for feed in list:
        o = OPML.Outline()
        o['text'] = feed.title.encode('utf-8')
        o['description'] = feed.channel_description.encode('utf-8')
        o['htmlUrl'] = feed.channel_link
        o['language'] = 'unknown'
        o['title'] = feed.channel_title.encode('utf-8')
        o['type'] = 'rss'
        o['version'] = 'RSS'
        o['xmlUrl'] = feed.access_info[0]
        opml.outlines.append(o)
    f = gnomevfs.create(fname, gnomevfs.OPEN_WRITE, 0)
    f.write('<?xml version="1.0"?>\n')
    opml.output(f)
    f.close()





""" Feed.py

Provides a module for managing feeds
"""
__copyright__ = "Copyright (c) 2002-2005 Free Software Foundation, Inc."
__license__ = """
Straw is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

Straw is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA. """


import Event
import FeedItems
import FeedDataRouter
import Config
from error import log

class Feed(object, Event.SignalEmitter):
    "A Feed object stores information set by user about a RSS feed."

    DEFAULT = -1
    STATUS_IDLE = 0
    STATUS_POLLING = 1

    __slots__ = ('_title', '_location', '_username', '_password', '_parsed',
                 '__save_fields', '_items', '_slots',
                 '_id', '_channel_description',
                 '_channel_title', '_channel_link', '_channel_copyright',
                 'channel_lbd', 'channel_editor', 'channel_webmaster',
                 'channel_creator', '_previous_etag',
                 '_error', '_process_status', 'router', 'sticky', '_parent',
                 '_items_stored', '_poll_freq', '_last_poll', '_n_items_unrea')

    __save_fields = (('_title', ""), ('_location', ""), ('_username', ""),
                     ('_password', ""), ('_id', ""),
                     ('_channel_description', ""), ('_channel_title', ""),
                     ('_channel_link', ""), ('_channel_copyright', ""),
                     ('channel_creator', ""),
                     ('_previous_etag', ""), ('_error', None),
                     ('_items_stored', DEFAULT),
                     ('_poll_freq', DEFAULT),
                     ('_last_poll', 0),
                     ('_n_items_unread', 0))


    # use one of the factory functions below instead of this directly
    def __init__(self, title="", location="", username="", password=""):
        Event.SignalEmitter.__init__(self)
        self.initialize_slots(Event.FeedInfoUpdatedSignal,
                              #Event.FeedUnreadChangedSignal,
                              #Event.FeedNumOfItemsStoredChangedSignal,
                              #Event.FeedPollFrequencyChangedSignal,
                              Event.FeedLastPollChangedSignal,
                              Event.FeedStatusChangedSignal,
                              Event.FeedErrorStatusChangedSignal,
                              Event.FeedPolledSignal,
                              Event.NewItemsSignal,
                              Event.ItemsAddedSignal,
                              Event.ItemReadSignal,
                              Event.ItemStickySignal,
                              Event.ItemDeletedSignal,
                              Event.AllItemsReadSignal,
                              Event.RefreshFeedDisplaySignal)

        self._title = title
        self._channel_description = ""
        self._channel_title = ""
        self._channel_link = ""
        self._channel_copyright = ""
        self.channel_lbd = None
        self.channel_editor = ""
        self.channel_webmaster = ""
        self.channel_creator = ""
        self._location = location
        self._username = username
        self._password = password
        self._parsed = None
        self._items = FeedItems.FeedItems(self)
        self._previous_etag = None
        self._error = None
        self._process_status = self.STATUS_IDLE
        self.router = FeedDataRouter.FeedDataRouter(self)
        self._parent = None
        self._items_stored = Feed.DEFAULT
        self._poll_freq = Feed.DEFAULT
        self._last_poll = 0
        self._n_items_unread = 0
        return

    def __str__(self):
        return "Feed '%s' from %s" % (self._title, self._location)

    @property
    def id(self):
        return self._id

    @apply
    def parsed():
        doc = "A ParsedSummary object generated from the summary file"
        def fget(self):
            return self._parsed
        def fset(self, parsed):
            self._parsed = parsed
        return property(**locals())

    @apply
    def title():
        doc = "The title of this Feed (as defined by user)"
        def fget(self):
            text = ''
            if self._title:
                text = self._title
            return text
        def fset(self, title):
            if self._title != title:
                self._title = title
                self.emit_signal(Event.FeedInfoUpdatedSignal(self))
        return property(**locals())

    @apply
    def access_info():
        doc = "A tuple of location, username, password"
        def fget(self):
            return (self._location, self._username, self._password)
        def fset(self, (location,username,password)):
            self._location = location
            self._username = username
            self._password = password
            self.emit_signal(Event.FeedInfoUpdatedSignal(self))
        return property(**locals())

    @apply
    def location():
        doc = ""
        def fget(self):
            return self._location
        def fset(self, location):
            if self._location != location:
                self._location = location
                self.emit_signal(Event.FeedInfoUpdateSignal(self))
        return property(**locals())

    @apply
    def channel_title():
        doc = ""
        def fget(self):
            text = ''
            if self._channel_title:
                text = self._channel_title
            return text
        def fset(self, t):
            changed = self._channel_title != t
            self._channel_title = t
            if changed:
                self.emit_signal(Event.FeedInfoUpdatedSignal(self))
        return property(**locals())

    @apply
    def channel_description():
        doc = ""
        def fget(self):
            text = ''
            if self._channel_description:
                text = self._channel_description
            return text
        def fset(self, t):
            changed = self._channel_description != t
            self._channel_description = t
            if changed:
                self.emit_signal(Event.FeedInfoUpdatedSignal(self))
        return property(**locals())

    @apply
    def channel_link():
        doc = ""
        def fget(self):
            return self._channel_link
        def fset(self, t):
            changed = self._channel_link != t
            self._channel_link = t
            if changed:
                self.emit_signal(Event.FeedInfoUpdatedSignal(self))
        return property(**locals())

    @apply
    def channel_copyright():
        doc = ""
        def fget(self):
            return self._channel_copyright
        def fset(self, t):
            changed = self._channel_copyright != t
            self._channel_copyright = t
            if changed:
                self.emit_signal(Event.FeedInfoUpdatedSignal(self))
        return property(**locals())

    @apply
    def number_of_items_stored():
        doc = ""
        def fget(self):
            return self._items_stored
        def fset(self, num=None):
            if self._items_stored != num:
                self._items_stored = num
                self.emit_signal(Event.FeedNumOfItemsStoredChangedSignal(self))
        return property(**locals())

    @apply
    def poll_frequency():
        doc = ""
        def fget(self):
            return self._poll_freq
        def fset(self, freq):
            if self._poll_freq != freq:
                self._poll_freq = freq
                self.emit_signal(Event.FeedPollFrequencyChangedSignal(self))
        return property(**locals())

    @apply
    def last_poll():
        doc = ""
        def fget(self):
            return self._last_poll
        def fset(self, time):
            if self._last_poll != time:
                self._last_poll = time
                self.emit_signal(Event.FeedLastPollChangedSignal(self))
        return property(**locals())

    @apply
    def n_items_unread():
        doc = ""
        def fget(self):
            return self._n_items_unread
        def fset(self, n):
            if self._n_items_unread != n:
                self._n_items_unread = n
                self.emit_signal(Event.FeedUnreadChangedSignal(self))
        return property(**locals())

    @apply
    def previous_etag():
        doc = ""
        def fget(self):
            return self._previous_etag
        def fset(self, etag):
            self._previous_etag = etag
        return property(**locals())

    @apply
    def error():
        doc = ""
        def fget(self):
            return self._error
        def fset(self, error):
            if self._error != error:
                self._error = error
                self.emit_signal(Event.FeedErrorStatusChangedSignal(self))
        return property(**locals())

    @apply
    def process_status():
        doc = ""
        def fget(self):
            return self._process_status
        def fset(self, status):
            if status != self._process_status:
                self._process_status = status
                self.emit_signal(Event.FeedStatusChangedSignal(self))
        return property(**locals())

    @apply
    def parent():
        doc = ""
        def fget(self):
            return self._parent
        def fset(self, parent):
            self._parent = parent
        return property(**locals())

    @property
    def next_refresh(self):
        """ return the feed's next refresh (time)"""
        nr = None
        if self._poll_freq == self.DEFAULT:
            increment = Config.get_instance().poll_frequency
        else:
            increment = self._poll_freq
        if increment > 0:
            nr = self.last_poll + increment
        return nr

    def dump(self):
        fl = {}
        for f, default in self.__save_fields:
            fl[f] = self.__getattribute__(f)
        return fl

    def undump(self, dict):
        for f, default in self.__save_fields:
            self.__setattr__(f, dict.get(f, default))
        return

    def poll_done(self):
        self.emit_signal(Event.FeedPolledSignal(self))

    def signal_new_items(self, items):
        self.emit_signal(Event.NewItemsSignal(self, items))

    def signal_added_items(self, items):
        self.emit_signal(Event.ItemsAddedSignal(self, items))

    def signal_deleted_item(self, item):
        self.emit_signal(Event.ItemDeletedSignal(self, item))

    def signal_refresh_display(self):
        self.emit_signal(Event.RefreshFeedDisplaySignal(self))

    def signal_all_items_read(self, items):
        self.emit_signal(Event.AllItemsReadSignal(self, items))

    def forward_signal(self, signal):
        self.emit_signal(signal)

    def add_items(self, items):
        return self._items.add_items(items)

    def restore_items(self, items):
        return self._items.restore_items(items)

    def delete_all_items(self):
        self.signal_deleted_item(self.items)

    def get_item_index(self, item):
        return self._items.get_item_index(item)

    @property
    def items(self):
        return self._items.get_items()

    @property
    def number_of_unread(self):
        return self._n_items_unread

    @property
    def number_of_items(self):
        return self._items.number_of_items

    def mark_all_read(self):
        return self._items.mark_all_read()

    def load_contents(self):
        return self._items.load()

    def unload_contents(self):
        self._items.unload()

    @classmethod
    def create_new_feed(klass, title, location="", username="", password=""):
        f = klass()
        f._title = title
        f._location = location
        f._id = Config.get_instance().next_feed_id_seq()
        f._username = username
        f._password = password
        return f

    @classmethod
    def create_empty_feed(klass):
        f = klass()
        return f



import dbus
import dbus.service
import dbus.glib
from dbus import DBusException
import subscribe
import error
import Config


# Based on the values from NetworkManager/include/NetworkManager.h
# We only care about CONNECTED and DISCONNECTED at the moment.
NM_STATE_CONNECTED = 3
NM_STATE_DISCONNECTED = 4

class FeedReader(dbus.service.Object):
        service_name = "org.gnome.feed.Reader"
        object_path = "/org/gnome/feed/Reader"

        def __init__(self):
		try:
                        self._session_bus = dbus.SessionBus()
                        self._service = dbus.service.BusName(self.service_name, bus=self._session_bus)
                        dbus.service.Object.__init__(self, self._service, self.object_path)
		except Exception, e:
			error.log_exc("Error while initializing dbus connection %s" % e)

        @dbus.service.method("org.gnome.feed.Reader")
	def Subscribe(self, url):
                subscribe.show(url=url)
		return True

class NetworkListener:
        SERVICE_NAME = "org.freedesktop.NetworkManager"
        SERVICE_PATH = "/org/freedesktop/NetworkManager"

        def __init__(self):
                self._config = Config.get_instance()

        def set_state(self, state):
                if state == NM_STATE_CONNECTED:
                        self._config.offline = False
                else:
                        self._config.offline = True

        def active_cb(self, path):
                self._config.offline = False

        def inactive_cb(self, path):
                self._config.offline = True


def start_services():
        fr = FeedReader()

        try:
                systemBus = dbus.SystemBus()
                proxy_obj = systemBus.get_object(NetworkListener.SERVICE_NAME,
                                                 NetworkListener.SERVICE_PATH)

                nl = NetworkListener()
                # don't touch offline if it has been previously set.
                if not Config.get_instance().offline:
                        nl.set_state(proxy_obj.state())

                nm_interface = dbus.Interface(proxy_obj, NetworkListener.SERVICE_NAME)
                nm_interface.connect_to_signal('DeviceNowActive', nl.active_cb)
                nm_interface.connect_to_signal('DeviceNoLongerActive', nl.inactive_cb)
        except DBusException, de:
                error.log("Error while connecting to NetworkManager: %s" % str(de))



import sys
import unittest
sys.path.insert(0, '../src/lib')
import utils

class UtilsTestCase(unittest.TestCase):

    def testFormatDateBadEncoding(self):
        # Bugzilla 145513
        encoding = "en_US.ISO8859-1"
        format = "%A"
        time = (2004, 8, 5, 20, 29, 59, 3, 218, 0)
        try:
            utils.format_date(time, format, encoding)
        except Exception, ex:
            self.fail("format_date: %s" % str(ex))

    def testURLlocation(self):
        url = "http://www.amazon.com/gp/aws/sdk/103-9053546-3314261"
        self.assertEqual("http://www.amazon.com", utils.get_url_location(url))

    def testReadText(self):
        text = "Blah &gt;foo&lt; <p>bar</p>"
        t = utils.read_text(text, 60)
        self.assertEqual(t, "Blah >foo< bar")

    def testCompleteUrl(self):
        tlist = [("images/foo.jpg","http://blah.com","http://blah.com/images/foo.jpg"),
                 ("/images/foo.jpg","http://b.com","http://b.com/images/foo.jpg")]
        for u,b,r in tlist:
            self.assertEqual(r,utils.complete_url(u,b))

def suite():
    suite = unittest.makeSuite(UtilsTestCase, 'test')
    return suite

if __name__ == '__main__':
    unittest.main()



""" ItemView.py

Module for displaying an item to the user
"""
__copyright__ = "Copyright (c) 2002-2005 Free Software Foundation, Inc."
__license__ = """
Straw is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

Straw is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA. """


import os
import re
import urlparse
from xml.sax import saxutils
import codecs
import pygtk
pygtk.require('2.0')
import gtk
import gtk.glade
import gtkhtml2
import utils
import dialogs
import MessageManager
import FeedCategoryList
import ImageCache
import MVP
import Config
import error
import dialogs

class HTMLView(MVP.View):
    """
    Widget: gtkhtml2.View
    Model: gtkhtml.Document
    Presenter: HTMLPresenter
    """
    ui = """
    <ui>
        <popup name=\"htmlview_popup\">
            <menuitem action=\"open_link_location\"/>
            <menuitem action=\"copy_link_location\"/>
            <menuitem action=\"copy_text\"/>
            <menuitem action=\"subscribe\"/>
            <menuitem action=\"zoom_in\"/>
            <menuitem action=\"zoom_out\"/>
            <menuitem action=\"zoom_100\"/>
        </popup>
    </ui>
    """

    def _initialize(self):
        self._widget.set_document(self._model)
        # Make the article view focusable. gtkhtml2.View scrolls to the first
        # link of the document which sometimes makes the first part of the
        # document unviewable if the first link is found at the bottom part of
        # the document. This is good for now since the template includes a
        # link at the topmost page of the view.
        self._widget.set_property('can-focus', True)
        self._widget.get_vadjustment().set_value(0)
        self._widget.connect('button_press_event', self._button_press_event)
        self._css = None
        self._url = None
        self._text_selection = None
        self._popup = None
        self._init_css()
        self._create_popup()


    def _connect_signals(self, widget):
        " We don't 'auto-connect' since self._widget is not a glade file"
        pass

    @property
    def widget(self):
        return self._widget

    def _button_press_event(self, widget, event):
        if event.button == 1:
            if self._url:
                self._presenter.display_url(self._url)
            self._url = self._text_selection = None
        elif event.button == 3:
            self._text_selection = self._presenter.get_html_text_selection()
            action_group = self._uimanager.get_action_groups()[-1]
            link_action = action_group.get_action('copy_link_location')
            open_link_action = action_group.get_action('open_link_location')
            text_action = action_group.get_action('copy_text')
            subs_action = action_group.get_action('subscribe')
            if self._url and self._text_selection:
                link_action.set_visible(True)
                open_link_action.set_visible(True)
                subs_action.set_visible(True)
                text_action.set_visible(True)
            elif self._url:
                text_action.set_visible(False)
                link_action.set_visible(True)
                open_link_action.set_visible(True)
                subs_action.set_visible(True)
            elif self._text_selection:
                link_action.set_visible(False)
                open_link_action.set_visible(False)
                subs_action.set_visible(False)
                text_action.set_visible(True)
            else:
                link_action.set_visible(False)
                open_link_action.set_visible(False)
                subs_action.set_visible(False)
                text_action.set_visible(False)
            self._uimanager.ensure_update()
            self._popup.popup(None, None, None, event.button,
                              gtk.get_current_event_time())
            return True

    def _create_popup(self):
        actions = [
            ("copy_text", gtk.STOCK_COPY, "_Copy", None, None,
             self._on_copy_text),
            ("copy_link_location", None, "_Copy Link Location",
             None, None, self._on_copy_url),
            ("open_link_location", None, "_Open Link",
             None, None, self._open_link_location),
            ("subscribe", None, "_Subscribe", None, None,
             self._subscribe),
            ("zoom_in", gtk.STOCK_ZOOM_IN, "Zoom _In", None, None,
             lambda *args: self._on_magnify("in")),
            ("zoom_out", gtk.STOCK_ZOOM_OUT, "Zoom _Out", None, None,
             lambda *args: self._on_magnify("out")),
            ("zoom_100", gtk.STOCK_ZOOM_100, "_Normal Size", None, None,
             lambda *args: self._on_magnify("reset"))
            ]
        self._uimanager = gtk.UIManager()
        actiongroup = gtk.ActionGroup("HtmlViewActions")
        actiongroup.add_actions(actions)
        self._uimanager.insert_action_group(actiongroup, -1)
        self._uimanager.add_ui_from_string(self.ui)
        self._popup = self._uimanager.get_widget("/htmlview_popup")


    def _on_magnify(self, action):
        if action == "in":
            self._widget.zoom_in()
        elif action == "out":
            self._widget.zoom_out()
        else:
            self._widget.zoom_reset()
        config = Config.get_instance()
        config.text_magnification = self._widget.get_magnification()

    def _on_copy_text(self, *args):
        self._presenter.set_clipboard_text(self._text_selection)
        gtkhtml2.html_selection_clear(self._widget)
        return

    def _on_copy_url(self, *args):
        self._presenter.set_clipboard_text(self._url)
        return

    def _open_link_location(self, *args):
        utils.url_show(self._url)

    def _subscribe(self, *args):
        import subscribe
        subscribe.show(url=self._url)
        return


    def _init_css(self):
        if self._css is None:
            css = file(os.path.join(
                utils.find_data_dir(), "straw.css")).read()
            # derive colors for blockquotes and header boxes from
            # the GTK+ theme
            # the GTK+ colors are in the range 0-65535
            bgdivisor = int(65535/(9.0/10))
            fgdivisor = 65535
            borderdivisor = int(65535/(6.0/10))
            gtkstyle = self._widget.get_style()

            headerbg = "background-color: #%.2x%.2x%.2x;" % (
            	(gtkstyle.bg[gtk.STATE_NORMAL].red * 255) / bgdivisor,
              	(gtkstyle.bg[gtk.STATE_NORMAL].blue * 255) / bgdivisor,
              	(gtkstyle.bg[gtk.STATE_NORMAL].green * 255) / bgdivisor)

            headerfg = "color: #%.2x%.2x%.2x;" % (
            	(gtkstyle.fg[gtk.STATE_NORMAL].red * 255) / fgdivisor,
              	(gtkstyle.fg[gtk.STATE_NORMAL].blue * 255) / fgdivisor,
              	(gtkstyle.fg[gtk.STATE_NORMAL].green * 255) / fgdivisor)

            headerborder = "border-color: #%.2x%.2x%.2x;" % (
            	(gtkstyle.bg[gtk.STATE_NORMAL].red * 255) / borderdivisor,
              	(gtkstyle.bg[gtk.STATE_NORMAL].blue * 255) / borderdivisor,
              	(gtkstyle.bg[gtk.STATE_NORMAL].green * 255) / borderdivisor)


            css = re.sub(r"/\*\*\*HEADERBG\*/", headerbg, css)
            css = re.sub(r"/\*\*\*HEADERFG\*/", headerfg, css)
            css = re.sub(r"/\*\*\*HEADERBORDERCOLOR\*/",
                               headerborder, css)

            css = re.sub(r"/\*\*\*BQUOTEBG\*/", headerbg, css)
            css = re.sub(r"/\*\*\*BQUOTEFG\*/", headerfg, css)
            css = re.sub(r"/\*\*\*BQUOTEBORDERCOLOR\*/",
                               headerborder, css)
            self._css = css
        return

    def report_error(self, title, description):
        dialogs.report_error(title, description)

    def get_css(self):
        return self._css

    def get_adjustments(self):
        return (self._widget.get_vadjustment(), self._widget.get_hadjustment())

    def get_widget(self):
        return self._widget

    def connect_widget_signal(self, signal, callback):
        self._widget.connect(signal, callback)

    def set_on_url(self, url):
        self._url = url

    def set_magnification(self, size):
        self._widget.set_magnification(size)

class HTMLPresenter(MVP.BasicPresenter):
    """
    Model: gtkhtml2.Document
    View: HTMLView
    """
    def _initialize(self):
        self._model.connect('request-url', self._request_url)
        self._view.connect_widget_signal('on_url', self._on_url)
        self._item = None

    def _on_url(self, view, url):
        self._view.set_on_url(url)
        if url:
            url = utils.complete_url(url, self._item.feed.location)
        else:
            url = ""
        mmgr = MessageManager.get_instance()
        mmgr.post_message(url)
        return

    def _request_url(self, document, url, stream):
        feed = self._item.feed
        try:
            try:
                url = utils.complete_url(url, self._item.feed.location)
                if urlparse.urlparse(url)[0] == 'file':
                    # local URL, don't use the cache.
                    f = file(urlparse.urlparse(url)[2])
                    stream.write(f.read())
                    f.close()
                else:
                    image = ImageCache.cache[url]
                    stream.write(image.get_data())
            except Exception, ex:
                error.log("Error reading image in %s: %s" % (url, ex))
        finally:
            stream.close()
            stream = None
        return

    def set_clipboard_text(self, text):
        utils.set_clipboard_text(text)

    def get_html_text_selection(self):
        return gtkhtml2.html_selection_get_text(self.view.widget)

    def display_url(self, link):
        link = link.strip()
        link = utils.complete_url(link, self._item.feed.location)
        try:
            utils.url_show(link)
        except Exception, ex:
            self._view.report_error(_("Error Loading Browser"),
                                    _("Please check your browser settings and try again."))
        return

    def get_view_adjustments(self):
        return self._view.get_adjustments()

    def get_view_widget(self):
        return self._view.get_widget()

    def display_item(self, item, encoding):
        self._item = item
        content = self._htmlify_item(item, encoding)
        self._prepare_stream(content)
        return

    def display_empty_feed(self):
        content = """<p class=\"emptyfeed\"/>"""# _("No data yet, need to poll first.") </p>"""
        self._prepare_stream(content)

    def display_empty_search(self):
        content = """
        <h2>Search Subscriptions</h2>
        <p>
        Begin searching by typing your text on the text box on the left side.
        </p>
        """
        self._prepare_stream(content)
        return

    def set_view_magnification(self, size):
        self.view.set_magnification(size)

    def _encode_for_html(self, unicode_data, encoding='utf-8'):
        """ From Python Cookbook, 2/ed, section 1.23
        'html_replace' is in the utils module
        """
        return unicode_data.encode(encoding, 'html_replace')

    def _prepare_stream(self, content):
        html = self._generate_html(content)
        html = self._encode_for_html(html)
        self._model.clear()
        self._model.open_stream("text/html")
        self._model.write_stream(html)
        self._model.close_stream()
        return

    def _generate_html(self, body):
        # heading
        html = """<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
        <html>
        <head><title>title</title>
        <meta http-equiv="Content-Type" content="text/html; charset=utf-8" />"""

        # stylesheet
        if Config.get_instance().reload_css:
            html += """<link rel="stylesheet" type="text/css" href="file://"
            """ + os.path.join(utils.find_data_dir(), "straw.css") + """/>"""
        else:
            html += """<style type="text/css">""" + self._view.get_css() + """</style>"""

        # body
        html += "</head><body>%s</body></html>" % body
        return html

    def _htmlify_item(self, item, encoding):
        feed = item.feed
        ret = []

        # item header
        ret.append('<div id="itemheader">')
        if item.title is not None:
            if item.link is not None:
                ret.append('<div class="title"><a href="%s">%s</a></div>' % (item.link,item.title))
            else:
                ret.append(item.title)
        ret.append('<table id="itemhead" cellspacing="0" cellpadding="0">')
        if item.pub_date is not None:
            timestr = utils.format_date(
                item.pub_date, utils.get_date_format(), encoding)
            ret.append(''.join(('<tr><td class="headleft" id="date">%s</td><td class="headright"></td></tr>' % str(timestr))))

        ret.append('</table>')
        ret.append('</div>')

        # item body
        if item.description is not None:
            item.description.replace('\n', '<br/>')
            ret.append('<div class="description">%s</div>' % item.description)

        if item.publication_name is not None:
            ret.append('<div class="description">')
            ret.append('<b>%s:</b> %s<br/>' % (_("Publication"),
                                               item.publication_name))
            if item.publication_volume is not None:
                ret.append('<b>%s:</b> %s ' % (_("Volume"),
                                                    item.publication_volume))
                if item.publication_number is not None:
                    ret.append('( %s )<br />' % item.publication_number)
            if item.publication_section is not None:
                ret.append('<b>%s:</b> %s<br />' % (_("Section"),
                                                    item.publication_section))
            if item.publication_starting_page is not None:
                ret.append('<b>%s:</b> %s' % (_("Starting Page"),
                                              item.publication_starting_page))
            ret.append('</div>')

        # freshmeat fields
        freshmeat_data = []
        if item.fm_license != '' and item.fm_license is not None:
            freshmeat_data.append('<p><b>%s:</b> %s</p>' %
                                  (_("Software license"), item.fm_license))
        if item.fm_changes != '' and item.fm_changes is not None:
            freshmeat_data.append('<p><b>%s:</b> %s</p>' %
                                  (_("Changes"), item.fm_changes))
        if len(freshmeat_data) > 0:
            ret.append('<div class="description">')
            ret.extend(freshmeat_data)
            ret.append('</div>')
        # empty paragraph to make sure that we get space here
        ret.append('<p></p>')
        # Additional information
        dcret = []
        
        # RSS Enclosures
        
        if item.enclosures:
            dcret.append('<tr class="tr.dc"><td class="dcname"><span>%s</span></td><td class="dcvalue"><table>' % _("Enclosed Media"))
            for enc in item.enclosures: # rss 2.0 defines only one enclosure per item
                size = int(enc.length)
                unit = _('bytes')
                if size > 1024:
                    unit = _('KB')
                    size /= 1024.0
                if size > 1024:
                    unit = _('MB')
                    size /= 1024.0
                link_text = enc['href'].split('/')[-1]
                
                # find what kind of media is that. enc[type] will have something like audio/mp3 or video/mpeg (mimetypes)
                # some symlinks are not present on the tango icon theme mimetype dir. audio and application are 2 good examples. So I am not relying on the symlinks now...
                kind = enc['type'].split('/')[0]
                if kind == 'audio':
                    icon_name = 'audio-x-generic'
                elif kind == 'video':
                    icon_name = 'video-x-generic'
                elif kind == 'image':
                    icon_name = 'image-x-generic'
                elif kind == 'application':
                    icon_name = 'binary'
                elif kind == 'text':
                    icon_name = 'text-x-generic'
                else:
                    icon_name = "unknown"
                
                it = gtk.icon_theme_get_default()
                ii = it.lookup_icon(icon_name, 32, gtk.ICON_LOOKUP_NO_SVG)
                imgsrc = 'file://' + ii.get_filename()
                dcret.append('<tr><td><div style="vertical-align: middle"><a class="dclink" href="%s" style="vertical-align: middle"><img style="padding: 0px 15px 5px 0px" src=%s /> %s</a> (%.2f %s - %s)</div></td></tr>' % (enc['href'], imgsrc, link_text, size, unit, enc['type']))
            dcret.append('</table></td></tr>')
        
        if item.creator is not None:
            dcret.append('<tr class="tr.dc"><td class="dcname"><span>%s</span></td><td class="dcvalue"><span>%s</span></td></tr>' % (_("Posted by"), item.creator))
        if item.contributors is not None and len(item.contributors):
            for c in item.contributors:
                dcret.append('<tr class="tr.dc"><td class="dcname"><span>%s</span></td><td class="dcvalue"><span>%s</span></td></tr>' \
                             % (_("Contributor:"), c.name))
        if item.source:
            url = utils.get_url_location(item.source['url'])
            text = saxutils.escape(url)
            dcret.append('<tr class="tr.dc"><td class="dcname"><span>%s</span></td><td class="dcvalue"><a class="dclink" href="%s"><span>%s</span></a></td></tr>' %
                         (_("Item Source"), url, text))

        if item.guid is not None and item.guid != "" and item.guidislink:
            dcret.append('<tr class="tr.dc"><td class="dcname"><span>%s</span></td><td class="dcvalue"><a class="dclink" href="%s"><span>%s</span></a></td></tr>' % (_("Permalink"), item.guid, item.guid))
        # check for not guidislink for the case where there is guid but
        # isPermalink="false" and yet link is the same as guid (link is
        # always assumed to be a valid link)
        if item.link != "" and item.link is not None and (item.link != item.guid or not item.guidislink):
             dcret.append('<tr class="tr.dc"><td class="dcname"><span>%s</span></td><td class="dcvalue"><a class="dclink" href="%s"><span>%s</span></a></td></tr>' %
                        (_("Complete story"), item.link, item.link))

        if item.license_urls:
            for l in item.license_urls:
                if l:
                    dcret.append('<tr class="tr.dc"><td class="dcname"><span>%s</span></td><td class="dcvalue"><a class="dclink" href="%s"><span>%s</span></a></td></tr>' % (_("License"), l, l))

        if len(dcret):
            ret.append('<div class="dcinfo">%s<table class="dc" id="footer">' % _("Additional information"))
            ret.append("".join(dcret))
            ret.append('</table>')
        ret.append('</div>')
        return "".join(ret)

class ScrollView(MVP.WidgetView):
    """
    Widget: html_scrolled_window
    """
    def set_adjustments(self, vadjustment, hadjustment):
        self._widget.set_hadjustment(hadjustment)
        self._widget.set_vadjustment(vadjustment)
        return

    def add_child(self, widget):
        self._widget.add(widget)
        return

    def show(self):
        self._widget.show_all()
        return

    def adjust_vertical_adjustment(self):
        va = self._widget.get_vadjustment()
        va.set_value(va.lower)
        return

    def get_vadjustment(self):
        return self._widget.get_vadjustment()

class ScrollPresenter(MVP.BasicPresenter):
    """
    View: ScrollView
    """
    def set_view_adjustments(self, vadjustment, hadjustment):
        self._view.set_adjustments(vadjustment, hadjustment)
        return

    def update_view(self):
        self._view.adjust_vertical_adjustment()
        return

    def scroll_down(self):
        va = self._view.get_vadjustment()
        old_value = va.get_value()
        new_value = old_value + va.page_increment
        limit = va.upper - va.page_size
        if new_value > limit:
            new_value = limit
        va.set_value(new_value)
        return new_value > old_value

    def show_view(self):
        self._view.show()
        return

class ItemView:
    def __init__(self, item_view_container):
        self._encoding = utils.get_locale_encoding()
        widget_tree = gtk.glade.get_widget_tree(item_view_container)
        document = gtkhtml2.Document()
        widget = gtkhtml2.View()
        html_view = HTMLView(widget, document)
        self._html_presenter = HTMLPresenter(document, html_view)

        widget = widget_tree.get_widget('html_scrolled_window')
        scroll_view = ScrollView(widget)
        self._scroll_presenter = ScrollPresenter(view=scroll_view)

        vadj, hadj = self._html_presenter.get_view_adjustments()
        child = self._html_presenter.get_view_widget()
        self._scroll_presenter.set_view_adjustments(vadj, hadj)
        self._scroll_presenter.view.add_child(child)
        self._scroll_presenter.show_view()

        config = Config.get_instance()
        self._html_presenter.set_view_magnification(config.text_magnification)

    def item_selection_changed(self, signal):
        if signal.item:
            self._display_item(signal.item)

    def _display_item(self, item):
        self._html_presenter.display_item(item, self._encoding)
        self._scroll_presenter.update_view()

    def display_empty_feed(self):
        self._html_presenter.display_empty_feed()

    def display_empty_search(self):
        self._html_presenter.display_empty_search()

    def scroll_down(self):
        return self._scroll_presenter.scroll_down()

    def get_selected_text(self):
        return self._html_presenter.get_html_text_selection()




""" SummaryItem.py

"""
__copyright__ = "Copyright (c) 2002-2005 Free Software Foundation, Inc."
__license__ = """
Straw is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

Straw is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA. """

import time
import locale
import Event
import ImageCache


class SummaryItem(object, Event.SignalEmitter):
    __slots__ = ('title', 'link', 'description', 'guid', 'pub_date', 'source',
                 '_images', '_seen', '_id', 'feed', '_slots',
                 'fm_license', 'fm_changes', 'creator', 'license_urls',
                 '_searchable_fields', '_sticky', 'publication_name',
                 'publication_volume', 'publication_number',
                 'publication_section', 'publication_starting_page',
                 'guidislink', 'contributors', 'enclosures')

    _searchable_fields = ('title', 'description', 'fm_license', 'fm_changes',
                          'creator', 'contributors')

    def __init__(self):
        Event.SignalEmitter.__init__(self)
        self.initialize_slots(Event.ItemReadSignal, Event.ItemStickySignal)
        self.title = None
        self.link = None
        self.description = None
        self.guid = None
        self.guidislink = True
        self.pub_date = time.localtime()
        self.source = None
        self._images = {}
        self._seen = 0
        self._id = None
        self.feed = None
        self.fm_license = None
        self.fm_changes = None
        self.creator = None
        self.contributors = list()
        self.license_urls = list()
        self.publication_name = None
        self.publication_volume = None
        self.publication_number = None
        self.publication_section = None
        self.publication_starting_page = None
        self._sticky = 0
        self.enclosures = None

    def __eq__(self, item):
        r = item.title == self.title and item.description == self.description
        if r:
            return True
        elif item.guid and self.guid and item.guid == self.guid:
            return True
        return False

    @apply
    def seen():
        doc = ""
        def fget(self):
            return self._seen
        def fset(self, seen = True):
            if self._seen != seen:
                self._seen = seen
                self.emit_signal(Event.ItemReadSignal(self))
        return property(**locals())

    @apply
    def sticky():
        doc = ""
        def fget(self):
            return self._sticky
        def fset(self, sticky):
            if self._sticky != sticky:
                self._sticky = sticky
                self.emit_signal(Event.ItemStickySignal(self))
        return property(**locals())

    @apply
    def id():
        doc = "Item ID (internal)"
        def fget(self):
            return self._id
        def fset(self, iid):
            self._id = iid
        return property(**locals())

    def _add_image(self, image_name, restore):
        ImageCache.cache.add_refer(image_name, restore, self)
        self._images[image_name] = ImageCache.cache[image_name]

    def add_image(self, image_name):
        self._add_image(image_name, restore = False)

    def restore_image(self, image_name):
        self._add_image(image_name, restore = True)

    def set_image(self, image_name, image):
        self._images[image_name] = image

    def get_image(self, image_name):
        return self._images.get(image_name, None)

    def clear_images(self):
        self._images = {}

    def image_keys(self):
        return self._images.keys()

    def set_seen_quiet(self, seen = 1):
        old = self._seen
        self._seen = seen
        return seen != old

    def match(self, text):
        def try_field(field):
            return type(field) in (
                type(''), type(u'')) and field.lower().find(text.lower()) > -1
        for f in self._searchable_fields:
            try:
                if try_field(self.__getattribute__(f)):
                    return 1
            except AttributeError:
                pass
        return 0

    def clean_up(self):
        for image in self._images:
            ImageCache.cache.remove_refer(image)



""" OPMLImport.py
"""

__copyright__ = "Copyright (c) 2002-2005 Free Software Foundation, Inc."
__license__ = """
Straw is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

Straw is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA. """


import OPML
import Feed
import FeedList

class BlogListEntry(object):
    __slots__ = ('text', 'url')

def _find_entries(outline):
    entries = []
    for c in outline.children:
        entries += _find_entries(c)
    type = outline.get('type', '')
    text = outline.get('text', '')
    e = None
    if type == 'link':
        url = outline.get('url', '')
        if url != '':
            e = BlogListEntry()
            e.text = text
            e.url = url
    else:
        xmlurl = outline.get('xmlUrl', '')
        e = BlogListEntry()
        e.text = text
        if text == '':
            title = outline.get('title', '')
            if title == '':
                e = None
            e.text = title
        if e != None:
            if xmlurl != '':
                # there's something in xmlurl. There's a good chance that's
                # our feed's URL
                e.url = xmlurl
            else:
                htmlurl = outline.get('htmlUrl', '')
                if htmlurl != '':
                    # there's something in htmlurl, and xmlurl is empty. This
                    # might be our feed's URL.
                    e.url = htmlurl
                else:
                    # nothing else to try.
                    e = None
    if e is not None:
        entries[0:0] = [e]
    return entries

def find_entries(outlines):
    entries = []
    for o in outlines:
        entries += _find_entries(o)
    return entries

def read(stream):
    try:
        o = OPML.parse(stream)
    except ValueError:
        return None
    entries = find_entries(o.outlines)
    ret = list()
    edict = dict()
    # avoid duplicates.
    for e in entries:
        ek = (e.text, e.url)
        edict[ek] = edict.get(ek, 0) + 1
        if edict[ek] < 2:
            ret.append(e)
    return ret

def import_opml(filename,category=None):
    feedlist = FeedList.get_instance()
    fstream = open (filename)
    opml = read(fstream)
    if not opml:
        return
    feeds = [feed.access_info[0] for feed in feedlist]
    newitems = [Feed.Feed.create_new_feed(b.text, b.url) for b in opml if b.url not in feeds]
    feedlist.extend(category, newitems)
    return



""" QueueDict.py

Sorted container of SummaryItems
"""
__copyright__ = "Copyright (c) 2002-2005 Free Software Foundation, Inc."
__license__ = """GNU General Public License

This program is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA. """

from bisect import insort

class ItemQueue(dict):
    def __init__(self):
        dict.__init__(self)
        self._reverse = 0
        self._keys = list()

    def __delitem__(self, key):
        try:
            self._keys.remove(key)
            dict.__delitem__(self, key)
        except KeyError:
            raise "FeedQD: No Such Key %d" % key

    def __iter__(self):
        return self.iteritems()

    def __contains__(self, item):
        return item in self.itervalues()

    def __setitem__(self, key, value):
        self.append(value)

    def _set_reverse(self, reverse):
        self._reverse = reverse

    def _get_reverse(self):
        return self._reverse

    sort_order = property(_get_reverse, _set_reverse)

    def append(self, item):
        if item.id not in self._keys:
            insort(self._keys, item.id)
        dict.__setitem__(self, item.id, item)

    def replace(self, item):
        self.append(item)
        k = self.keys()
        item = None
        for iid in reversed(k):
            item = self[iid]
            if item.sticky:
                continue
            else:
                del self[iid]
                break
        return item

    def clear(self):
        dict.clear(self)
        del self._keys[:]

    def keys(self):
        if self._reverse:
            return sorted(self._keys, reverse=True)
        return self._keys

    def values(self):
        return map(self.get, self.keys())

    def iteritems(self):
        def sorteditems():
            for k in self.keys():
                yield (k, self[k])
        return sorteditems()

    def iterkeys(self):
        def sortedkeys():
            for k in self.keys():
                yield k
        return sortedkeys()

    def itervalues(self):
        def sortedvalues():
            for k in self.keys():
                yield self[k]
        return sortedvalues()

    def items(self):
        return [item for item in self.iteritems()]

    def index(self, key):
        try:
            idx = self.keys().index(key)
        except (IndexError, ValueError):
            idx = None
        return idx



""" CategorySelector.py

Module for selecting categories
"""
__copyright__ = "Copyright (c) 2002-2005 Free Software Foundation, Inc."
__license__ = """
Straw is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

Straw is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA. """

import locale
import pygtk
pygtk.require('2.0')
import gobject
import gtk
import FeedCategoryList
import MVP
import Event

class Column:
    TITLE=0
    CATEGORY=1

class CategoryView(MVP.WidgetView):
    """
    Widget: gtk.ComboBox
    Model: ListStore
    """
    def _initialize(self):
        self._popup = None
        self._widget.set_wrap_width(1)
        title = gtk.CellRendererText()
        self._widget.pack_start(title,False)
        self._widget.add_attribute(title,'text',Column.TITLE)
        self._widget.connect('changed', self._selection_changed)
        return

    def _model_set(self):
        self._widget.set_model(self._model)
        return

    def _selection_changed(self, combo, data=None):
        self._presenter.selection_changed(combo, data)

    def set_active(self, index): self._widget.set_active(index)
    def get_active(self): return self._widget.get_active()

    def get_active_iter(self): return self._widget.get_active_iter()
    def set_active_iter(self, it): self._widget.set_active_iter(it)

class CategoryPresenter(MVP.BasicPresenter):
    def _initialize(self):
        self.initialize_slots(Event.CategorySelectionChangedSignal)
        self.model = gtk.ListStore(gobject.TYPE_STRING, gobject.TYPE_PYOBJECT)
        self.model.set_sort_func(Column.TITLE, self._sort_func)
        self.model.set_sort_column_id(Column.TITLE, gtk.SORT_ASCENDING)
        self._init_signals()
        return

    def _sort_func(self, model, a, b):
        """
        From the gtk.TreeSortable.set_sort_func doc:

        The comparison callback should return -1 if the iter1 row should come before
        the iter2 row, 0 if the rows are equal, or 1 if the iter1 row should come
        after the iter2 row.
        """
        fclist = FeedCategoryList.get_instance()
        retval = 0
        cat_a = model.get_value(a, Column.CATEGORY)
        cat_b = model.get_value(b, Column.CATEGORY)

        if cat_a in fclist.pseudo_categories:
            retval = -1
        elif cat_b in fclist.pseudo_categories:
            retval = 1
        elif cat_a is not None and cat_b is not None:
            retval = locale.strcoll(cat_a.title, cat_b.title)
        elif cat_a is not None: retval = -1
        elif cat_b is not None: retval = 1
        return retval

    def _init_signals(self):
        fclist = FeedCategoryList.get_instance()
        fclist.signal_connect(Event.FeedCategoryListLoadedSignal,
                              self._display_categories)
        fclist.signal_connect(Event.FeedCategoryAddedSignal,
                              self._category_added_cb)
        fclist.signal_connect(Event.FeedCategoryRemovedSignal,
                              self._category_removed_cb)
        fclist.signal_connect(Event.FeedCategoryChangedSignal,
                              self._category_changed_cb)
        return

    def _get_selected_category(self):
        category = None
        if len(self._model):
            active = self._view.get_active()
            category = self._model[active][Column.CATEGORY]
        return category

    def _category_added_cb(self, signal):
        newcat = signal.category
        treeiter = self._model.append()
        self._model.set(treeiter,
                        Column.TITLE, newcat.title,
                        Column.CATEGORY, newcat)
        return

    def _category_removed_cb(self, signal):
        self._model.foreach(self._foreach_remove_cat,
                            signal.category)
        self._update_selection()
        return

    def _foreach_remove_cat(self, model, path, iter, category):
        cat = model[path][Column.CATEGORY]
        if cat is category:
            model.remove(iter)
            return True
        return False

    def _category_changed_cb(self, signal):
        self._model.foreach(self._foreach_category_changed,
                            signal.sender)
        self._update_selection()
        return

    def _foreach_category_changed(self, model, path, iter, category):
        cat = model[path][Column.CATEGORY]
        if cat is category:
            model.set(iter,
                      Column.TITLE, category.title,
                      Column.CATEGORY, category)
            return True
        return False

    def _display_categories(self, signal):
        fclist = signal.sender
        self._model.clear()
        for category in fclist.pseudo_categories:
            it = self._model.append()
            self._model.set(it,
                            Column.TITLE, category.title,
                            Column.CATEGORY, category)
        for category in fclist.user_categories:
            it = self._model.append()
            self._model.set(it,
                            Column.TITLE, category.title,
                            Column.CATEGORY, category)
        self._update_selection()
        return

    def _update_selection(self):
        active = self._view.get_active()
        if active < 0:
            self._view.set_active_iter(self._model.get_iter_first())
        return

    def selection_changed(self, combo, data=None):
        model = combo.get_model()
        newitem = model[combo.get_active()][1]
        self.emit_signal(Event.CategorySelectionChangedSignal(self,newitem))
        return

    def category_selected(self, category):
        self._model.foreach(self._foreach_category_selected,
                            category)
        self._update_selection()

    def _foreach_category_selected(self, model, path, iter, category):
        cat = model[path][Column.CATEGORY]
        if cat is category:
            self._view.set_active_iter(iter)
            return True
        return False



""" PreferencesDialog.py

Module setting user preferences.
"""
__copyright__ = "Copyright (c) 2002-2005 Free Software Foundation, Inc."
__license__ = """ GNU General Public License

Straw is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

Straw is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA. """

import pygtk
pygtk.require('2.0')
import gobject
import gtk
import pango
from gtk import glade
import locale
import error
import MVP
from weakref import WeakKeyDictionary
import Event
import FeedCategoryList
import FeedList
import Config
import ValueMonitor

class CategoryListView(MVP.WidgetView):
    """View object for the category tree view"""
    COLUMN_TITLE = 0
    COLUMN_OBJ = 1
    COLUMN_BOLD = 2
    COLUMN_EDITABLE = 3

    def _initialize(self):
        treemodel = gtk.ListStore(gobject.TYPE_STRING, gobject.TYPE_PYOBJECT,
                                  gobject.TYPE_INT, gobject.TYPE_BOOLEAN)
        treemodel.set_sort_func(self.COLUMN_TITLE, self._sort_feeds)
        treemodel.set_sort_column_id(self.COLUMN_TITLE, gtk.SORT_ASCENDING)
        self._widget.set_model(treemodel)
        self._widget.set_rules_hint(False)
        self._create_columns()
        self._widget.get_selection().connect(
            'changed', self._selection_changed, self.COLUMN_OBJ)

    def _sort_feeds(self, model, a, b):
        """
        From the gtk.TreeSortable.set_sort_func doc:

        The comparison callback should return -1 if the iter1 row should come before
        the iter2 row, 0 if the rows are equal, or 1 if the iter1 row should come
        after the iter2 row.
        """
        retval = 0
        feed_a = model.get_value(a, self.COLUMN_OBJ)
        feed_b = model.get_value(b, self.COLUMN_OBJ)

        if feed_a in self._model.pseudo_categories:
            retval = -1
        elif feed_b in self._model.pseudo_categories:
            retval = 1
        elif feed_a is not None and feed_b is not None:
            retval = locale.strcoll(feed_a.title, feed_b.title)
        elif feed_a is not None: retval = -1
        elif feed_b is not None: retval = 1
        return retval

    def _model_set(self):
        self._model.signal_connect(Event.FeedCategoryAddedSignal,
                                   self._category_added)
        self._model.signal_connect(Event.FeedCategoryRemovedSignal,
                                   self._category_removed)
        self._model.signal_connect(Event.FeedCategoryChangedSignal,
                                   self._category_changed)
        self._model.signal_connect(Event.FeedCategoryListLoadedSignal,
                                   self._categories_loaded)
        self._populate_treemodel()

    def _create_columns(self):
        renderer = gtk.CellRendererText()
        renderer.connect('edited', self._cell_edited)
        column = gtk.TreeViewColumn(_('_Category'), renderer,
                                    text=self.COLUMN_TITLE,
                                    weight=self.COLUMN_BOLD,
                                    editable=self.COLUMN_EDITABLE)
        self._widget.append_column(column)

    def _populate_treemodel(self):
        treemodel = self._widget.get_model()
        treemodel.clear()
        for category in self._model.pseudo_categories:
            iter = treemodel.append()
            treemodel.set(iter, self.COLUMN_TITLE, category.title,
                          self.COLUMN_OBJ, category,
                          self.COLUMN_BOLD, pango.WEIGHT_BOLD,
                          self.COLUMN_EDITABLE, False)
        for category in self._model.user_categories:
            iter = treemodel.append()
            treemodel.set(iter, self.COLUMN_TITLE, category.title,
                          self.COLUMN_OBJ, category,
                          self.COLUMN_BOLD, pango.WEIGHT_NORMAL,
                          self.COLUMN_EDITABLE, True)

    def _iterate_model_conditionally(self, func):
        treemodel = self._widget.get_model()
        iter = treemodel.get_iter_first()
        while iter is not None:
            if func(treemodel, iter): break
            iter = treemodel.iter_next(iter)

    def _add_category(self, category):
        self._populate_treemodel()
        def edit_row(treemodel, iter):
            if treemodel.get_value(iter, self.COLUMN_OBJ) is category:
                column = self._widget.get_column(self.COLUMN_TITLE)
                path = treemodel.get_path(iter)
                self._widget.set_cursor(path, column, True)
                return True
        self._iterate_model_conditionally(edit_row)

    def _remove_category(self, category):
        def remove_row(treemodel, iter):
            if treemodel.get_value(iter, self.COLUMN_OBJ) is category:
                column = self._widget.get_column(self.COLUMN_TITLE)
                iterpath = treemodel.get_path(iter)
                isIter = treemodel.remove(iter)
                if not isIter:
                    current = iterpath[:-1] + (iterpath[-1]-1,)
                else:
                    current = treemodel.get_path(iter)
                self._widget.set_cursor(current, column)
                return True
            return False
        self._widget.grab_focus()
        self._iterate_model_conditionally(remove_row)

    def _update_category_name(self, category):
        def update_title(treemodel, iter):
            if treemodel.get_value(iter, self.COLUMN_OBJ) is category:
                treemodel.set(iter, self.COLUMN_TITLE, category.title)
                return True
        self._iterate_model_conditionally(update_title)

    def _category_added(self, signal):
        self._add_category(signal.category)

    def _category_removed(self, signal):
        self._remove_category(signal.category)

    def _category_changed(self, signal):
        self._update_category_name(signal.sender)

    def _categories_loaded(self, signal):
        self._populate_treemodel()

    def _selection_changed(self, selection, *args):
        if self._presenter is None:
            error.log("presenter not set!")
            return
        model, iter = selection.get_selected()
        if iter is None:
            return
        category = model.get_value(iter, self.COLUMN_OBJ)
        self._presenter.category_changed(category)

    def _cell_edited(self, cell, path_string, text):
        treemodel = self._widget.get_model()
        iter = treemodel.get_iter_from_string(path_string)
        if not iter: return
        category = treemodel.get_value(iter, self.COLUMN_OBJ)
        self._presenter.category_title_edited(category, text)

class CategorySelectionChangedSignal(Event.BaseSignal):
    def __init__(self, sender, category):
        Event.BaseSignal.__init__(self, sender)
        self.category = category

class CategoryListPresenter(MVP.BasicPresenter):
    """Presenter object for the category tree view"""
    def _initialize(self):
        self.initialize_slots(CategorySelectionChangedSignal)
        self._selected_category = None

    def category_changed(self, category):
        self._selected_category = category
        self.emit_signal(CategorySelectionChangedSignal(self, category))

    def category_title_edited(self, category, title):
        category.title = title

    def add_category(self):
        cat = FeedCategoryList.FeedCategory(_("Category name"))
        self._model.add_category(cat)

    def remove_category(self):
        if self._selected_category is not None:
            self._model.remove_category(self._selected_category)

class CategoryAddView(MVP.WidgetView):
    """View object for the add category button"""
    def _on_category_add_button_clicked(self, *args):
        if self._presenter is not None:
            self._presenter.add_category()

class AddCategorySignal(Event.BaseSignal):
    pass

class CategoryAddPresenter(MVP.BasicPresenter):
    """Presenter object for the add category button"""
    def _initialize(self):
        self.initialize_slots(AddCategorySignal)

    def add_category(self):
        self.emit_signal(AddCategorySignal(self))

class CategoryRemoveView(MVP.WidgetView):
    """View object for the remove category button"""
    def _on_category_delete_button_clicked(self, *args):
        if self._presenter is not None:
            self._presenter.remove_category()

class RemoveCategorySignal(Event.BaseSignal):
    pass

class CategoryRemovePresenter(MVP.BasicPresenter):
    """Presenter object for the remove category button"""
    def _initialize(self):
        self.initialize_slots(RemoveCategorySignal)

    def remove_category(self):
        self.emit_signal(RemoveCategorySignal(self))

class CategoriesPresenter(object, Event.SignalEmitter):
    """Presenter object that combines the category list presenter and
    category add button presenter"""
    def __init__(self, lister, adder, remover):
        Event.SignalEmitter.__init__(self)
        self.initialize_slots(CategorySelectionChangedSignal)

        lister.signal_connect(CategorySelectionChangedSignal,
                              lambda s: self.emit_signal(s))
        adder.signal_connect(AddCategorySignal, self._add_category)
        remover.signal_connect(RemoveCategorySignal, self._remove_category)
        self._lister = lister
        self._adder = adder
        self._remover = remover

    def _add_category(self, signal):
        self._lister.add_category()

    def _remove_category(self, signal):
        self._lister.remove_category()
    
class FeedListView(MVP.WidgetView):
    """View object for the feed list"""
    COLUMN_TITLE = 0
    COLUMN_OBJ = 1
    COLUMN_CHECKED = 2
    COLUMN_TOGGLABLE = 3

    def _initialize(self):
        treemodel = gtk.ListStore(gobject.TYPE_STRING, gobject.TYPE_PYOBJECT,
                                  gobject.TYPE_BOOLEAN, gobject.TYPE_BOOLEAN)
        treemodel.set_sort_func(self.COLUMN_TITLE, self._sort_feeds)
        treemodel.set_sort_column_id(self.COLUMN_TITLE, gtk.SORT_ASCENDING)
        self._widget.set_model(treemodel)
        self._widget.set_rules_hint(False)
        self._create_columns()

        self._category = None
        self._updating_display = False

    def _model_set(self):
        self._model.signal_connect(Event.FeedsChangedSignal,
                                   self._model_updated)
        fclist = FeedCategoryList.get_instance()
        self._populate_treemodel()

    def _create_columns(self):
        renderer = gtk.CellRendererToggle()
        renderer.connect('toggled', self._cell_toggled)
        column = gtk.TreeViewColumn(_('_Member'), renderer,
                                    active=self.COLUMN_CHECKED,
                                    activatable=self.COLUMN_TOGGLABLE)
        column.set_sort_column_id(self.COLUMN_CHECKED)
        self._widget.append_column(column)
        renderer = gtk.CellRendererText()
        column = gtk.TreeViewColumn(_('_Feed'), renderer,
                                    text=self.COLUMN_TITLE)
        column.set_sort_column_id(self.COLUMN_TITLE)
        self._widget.append_column(column)

    def _populate_treemodel(self):
        treemodel = self._widget.get_model()
        treemodel.clear()
        feedlist = FeedList.get_instance()
        if self._category is None: return
        category_feeds = self._category.feeds
        enabletoggle = not isinstance(
            self._category, FeedCategoryList.PseudoCategory)
        for feed in feedlist:
            iter = treemodel.append()
            treemodel.set(iter, self.COLUMN_TITLE, feed.title,
                          self.COLUMN_OBJ, feed,
                          self.COLUMN_CHECKED, feed in category_feeds,
                          self.COLUMN_TOGGLABLE, enabletoggle)

    def _sort_feeds(self, model, a, b):
        feed_a = model.get_value(a, self.COLUMN_OBJ)
        feed_b = model.get_value(b, self.COLUMN_OBJ)

        if feed_a and feed_b :
            return locale.strcoll(feed_a.title.lower(), feed_b.title.lower())
        elif feed_a is not None: return -1
        elif feed_b is not None: return 1
        else:
            return 0

    def _model_updated(self, signal):
        self._populate_treemodel()

    def category_selected(self, category):
        self._category = category
        self._populate_treemodel()
        return

    def _cell_toggled(self, cell, path_string):
        if self._updating_display:
            return
        treemodel = self._widget.get_model()
        treeiter = treemodel.get_iter_from_string(path_string)
        if not treeiter: return
        feed = treemodel.get_value(treeiter, self.COLUMN_OBJ)
        self._presenter.feed_toggled(feed)
        self._populate_treemodel()

class FeedListPresenter(MVP.BasicPresenter):
    """Presenter object for the feed list"""
    def feed_toggled(self, feed):
        if self._category is None:
            return
        if feed in self._category.feeds:
            self._category.remove_feed(feed)
        else:
            self._category.append_feed(feed, False)

    def category_selected(self, category):
        self._category = category
        if self._view is not None:
            self._view.category_selected(category)

class CategoriesTab:
    def __init__(self, xml):
        self._selcategory = None # currently selected category

        self._delete_button = xml.get_widget('category_delete_button')
        self._delete_button.set_sensitive(False)

        self._selcategory = None

        clpresenter = CategoryListPresenter(
            model = FeedCategoryList.get_instance(),
            view = CategoryListView(xml.get_widget('category_treeview')))
        capresenter = CategoryAddPresenter(
            view = CategoryAddView(xml.get_widget('category_add_button')))
        crpresenter = CategoryRemovePresenter(
            view = CategoryRemoveView(
            xml.get_widget('category_delete_button')))

        self._categories_presenter = CategoriesPresenter(
            clpresenter, capresenter, crpresenter)
        self._categories_presenter.signal_connect(
            CategorySelectionChangedSignal, self._category_changed)

        feedmodel = FeedList.get_instance()
        self._feeds_presenter = FeedListPresenter()
        self._feeds_presenter.model = feedmodel
        self._feeds_presenter.view = FeedListView(
            xml.get_widget('feeds_treeview'))
        
        self._source_field = xml.get_widget(
            'category_external_source')
        self._source_username_field = xml.get_widget(
            'category_external_source_username')
        self._source_password_field = xml.get_widget(
            'category_external_source_password')
        self._source_frequency_spin = xml.get_widget(
            'category_external_source_frequency_spin')
        self._source_frequency_default_check = xml.get_widget(
            'category_refresh_default_check')
        
        sizegroup = gtk.SizeGroup(gtk.SIZE_GROUP_HORIZONTAL)
        for l in [xml.get_widget(w) for w in (
            'category_external_source_label',
            'category_external_source_username_label',
            'category_external_source_password_label',
            'category_external_source_frequency_label')]:
            sizegroup.add_widget(l)
        
        self._category_source_monitors = CategorySourceMonitors()

        nameFuncMap = {}
        for key in dir(self.__class__):
            if key[:4] == '_on_':
                nameFuncMap[key[1:]] = getattr(self, key)
        xml.signal_autoconnect(nameFuncMap)

        self._switching_categories = False

    def _setup_category_source_monitors(self, oldcat, newcat):
        self._category_source_monitors.disconnect(oldcat)
        
        subscription_set = newcat.subscription is not None
        is_pseudo = isinstance(newcat, FeedCategoryList.PseudoCategory)
        default_refresh = (
            (not subscription_set) or
            (newcat.subscription.frequency ==
             newcat.subscription.REFRESH_DEFAULT))
        attrfields = (
            ((subscription_set and newcat.subscription.location) or "",
             self._source_field),
            ((subscription_set and newcat.subscription.username) or "",
             self._source_username_field),
            ((subscription_set and newcat.subscription.password) or "",
             self._source_password_field))

        for f, w in attrfields:
            w.set_text(f)
            w.set_sensitive(not is_pseudo)
        if default_refresh:
            self._source_frequency_spin.set_value(
                Config.get_instance().poll_frequency / 60)
        else:
            self._source_frequency_spin.set_value(
                (subscription_set and newcat.subscription.frequency / 60) or 0)
        
        self._source_frequency_spin.set_sensitive(
            (not is_pseudo) and (not default_refresh))
        self._source_frequency_default_check.set_active(
            (not is_pseudo) and default_refresh)

        self._source_frequency_default_check.set_sensitive(not is_pseudo)

        def save_location(value, widget): newcat.subscription.location = value
        def save_username(value, widget): newcat.subscription.username = value
        def save_password(value, widget): newcat.subscription.password = value
        def save_frequency(value, widget):
            newcat.subscription.frequency = value * 60

        def create_sub_ensurer(saver):
            def ensure_and_save(value, widget):
                if newcat.subscription is None:
                    newcat.subscription = FeedCategoryList.OPMLCategorySubscription()
                saver(value, widget)
            return ensure_and_save
        
        if not self._category_source_monitors.has_key(newcat):
            monitors = dict(
                [(key, ValueMonitor.create_for_gtkentry(
                v, 5000, create_sub_ensurer(f)))
                for key, v, f in (("location", self._source_field,
                                   save_location),
                                  ("username", self._source_username_field,
                                   save_username),
                                  ("password", self._source_password_field,
                                   save_password))])
            monitors.update(
                {"frequency": ValueMonitor.create_for_gtkspin(
                self._source_frequency_spin, 5000,
                create_sub_ensurer(save_frequency))})
            self._category_source_monitors[newcat] = monitors

        self._category_source_monitors.connect(newcat)

    def _category_changed(self, signal):
        self._switching_categories = True
        category = signal.category
        self._setup_category_source_monitors(
            self._selcategory, category)
        self._selcategory = category

        if isinstance(category, FeedCategoryList.PseudoCategory):
            self._delete_button.set_sensitive(False)
        else:
            self._delete_button.set_sensitive(True)
        self._feeds_presenter.category_selected(category)
        self._switching_categories = False
        return

    def _on_category_sort_ascending_button_clicked(self, widget):
        self._selcategory.sort()

    def _on_category_sort_descending_button_clicked(self, widget):
        self._selcategory.sort()
        self._selcategory.reverse()

    def _on_category_refresh_default_check_toggled(self, widget):
        if self._switching_categories:
            return
        default_frequency = Config.get_instance().poll_frequency / 60
        if self._selcategory.subscription is None:
            self._selcategory.subscription = FeedCategoryList.OPMLCategorySubscription()
        if widget.get_active():
            self._source_frequency_spin.set_value(default_frequency)
            self._source_frequency_spin.set_sensitive(False)
            # flush the monitor so we don't get pending updates to the
            # frequency overriding this change
            self._category_source_monitors.flush_monitor(
                self._selcategory, "frequency")
            self._selcategory.subscription.frequency = self._selcategory.subscription.REFRESH_DEFAULT
        else:
            self._source_frequency_spin.set_sensitive(True)
            self._selcategory.subscription.frequency = default_frequency / 60


class PreferencesDialog:
    SEC_PER_MINUTE = 60

    def __init__(self, xml, parent):
        config = Config.get_instance()
        self._window = xml.get_widget('preferences_dialog')
        self._window.set_transient_for(parent)
        self._browser_override = xml.get_widget("prefs_browser_setting")
        self._browser_entry = xml.get_widget("prefs_browser_setting_entry")

        # General
        config = Config.get_instance()
        poll_frequency = int(config.poll_frequency/self.SEC_PER_MINUTE)
        items_stored = int(config.number_of_items_stored)

        browser_cmd = str (config.browser_cmd)
        if browser_cmd:
            self._browser_override.set_active(True)
            self._browser_entry.set_text (browser_cmd)
            self._browser_entry.set_sensitive (True)
        else:
            self._browser_entry.set_text (_("Using desktop setting"))

        xml.get_widget('poll_frequency_spin').set_value(poll_frequency)
        xml.get_widget('number_of_items_spin').set_value(items_stored)
        xml.get_widget(['item_order_oldest',
                        'item_order_newest'][config.item_order]).set_active(1)

        nameFuncMap = {}
        for key in dir(self.__class__):
            if key[:3] == 'on_':
                nameFuncMap[key] = getattr(self, key)
        xml.signal_autoconnect(nameFuncMap)

        self._categories_tab = CategoriesTab(glade.get_widget_tree(
            xml.get_widget('preferences_categories_tab')))

    def show(self, *args):
        self._window.present()

    def hide(self, *args):
        self._window.hide()

    def on_preferences_dialog_delete_event(self, *args):
        self.hide()
        return True

    def on_preferences_close_button_clicked(self, button):
        self.hide()
        return

    def on_poll_frequency_spin_value_changed(self, spin):
        Config.get_instance().poll_frequency = int(spin.get_value() * self.SEC_PER_MINUTE)

    def on_number_of_items_spin_value_changed(self, spin):
        Config.get_instance().number_of_items_stored = int(spin.get_value())

    def on_item_order_newest_toggled(self, radio):
        config = Config.get_instance()
        config.item_order = not config.item_order

    def on_item_order_oldest_toggled(self, radio):
        pass

    def on_prefs_browser_setting_toggled(self, button):
        active = button.get_active()
        if not active:
            config = Config.get_instance()
            config.browser_cmd = ""
        self._browser_entry.set_sensitive(active)

    def on_prefs_browser_setting_entry_focus_out_event(self, entry, *args):
        config = Config.get_instance()
        config.browser_cmd = entry.get_text()

class CategorySourceMonitors(WeakKeyDictionary):
    def disconnect(self, category):
        if self.has_key(category):
            for m in self[category].values():
                m.disconnect()

    def connect(self, category):
        if self.has_key(category):
            for m in self[category].values():
                m.connect()

    def flush_monitor(self, category, monitor):
        if self.has_key(category) and self[category].has_key(monitor):
            self[category][monitor].flush()



"""Ultra-liberal feed finder

http://diveintomark.org/projects/feed_finder/

Usage:
getFeeds(uri) - returns list of feeds associated with this address

Example:
>>> import feedfinder
>>> feedfinder.getFeeds('http://diveintomark.org/')
['http://diveintomark.org/xml/atom.xml']
>>> feedfinder.getFeeds('macnn.com')
['http://www.macnn.com/macnn.rdf']

Can also use from the command line.  Feeds are returned one per line:
$ python feedfinder.py diveintomark.org
http://diveintomark.org/xml/atom.xml

How it works:
0. At every step, feeds are minimally verified to make sure they are really feeds.
1. If the URI points to a feed, it is simply returned; otherwise
   the page is downloaded and the real fun begins.
2. Feeds pointed to by LINK tags in the header of the page (autodiscovery)
3. <A> links to feeds on the same server ending in ".rss", ".rdf", ".xml", or ".atom"
4. <A> links to feeds on the same server containing "rss", "rdf", "xml", or "atom"
5. <A> links to feeds on external servers ending in ".rss", ".rdf", ".xml", or ".atom"
6. <A> links to feeds on external servers containing "rss", "rdf", "xml", or "atom"
7. As a last ditch effort, we search Syndic8 for feeds matching the URI

"""

__version__ = "1.2"
__date__ = "2004-01-09"
__author__ = "Mark Pilgrim (f8dy@diveintomark.org)"
__copyright__ = "Copyright 2002-4, Mark Pilgrim"
__license__ = "Python"
__credits__ = """Abe Fettig for a patch to sort Syndic8 feeds by popularity
Also Jason Diamond, Brian Lalor for bug reporting and patches"""
__history__ = """
1.1 - MAP - 2003/02/20 - added support for Robot Exclusion Standard.  Will
fetch /robots.txt once per domain and verify that URLs are allowed to be
downloaded.  Identifies itself as
  rssfinder/<version> Python-urllib/<version> +http://diveintomark.org/projects/rss_finder/
1.2 - MAP - 2004-01-09 - added Atom support, changed name, relicensed,
  don't query Syndic8 by default (pass querySyndic8=1 to getFeeds to do it anyway)
"""

_debug = 0

# ---------- required modules (should come with any Python distribution) ----------
import sgmllib, urllib, urlparse, re, sys, robotparser

# ---------- optional modules (feedfinder will work without these, but with reduced functionality) ----------

# timeoutsocket allows feedfinder to time out rather than hang forever on ultra-slow servers.
# Python 2.3 now has this functionality available in the standard socket library, so under
# 2.3 you don't need to install anything.
import socket
if hasattr(socket, 'setdefaulttimeout'):
    socket.setdefaulttimeout(10)
else:
    try:
        import timeoutsocket # http://www.timo-tasi.org/python/timeoutsocket.py
        timeoutsocket.setDefaultSocketTimeout(10)
    except ImportError:
        pass

# XML-RPC support allows feedfinder to query Syndic8 for possible matches.
# Python 2.3 now comes with this module by default, otherwise you can download it
try:
    import xmlrpclib # http://www.pythonware.com/products/xmlrpc/
except ImportError:
    xmlrpclib = None

if not dict:
    def dict(aList):
        rc = {}
        for k, v in aList:
            rc[k] = v
        return rc
    
def _debuglog(message):
    if _debug: print message

class RobotFileParserFixed(robotparser.RobotFileParser):
    """patched version of RobotFileParser, integrating fixes from Python 2.3a2 and bug 690214"""
    
    def can_fetch(self, useragent, url):
        """using the parsed robots.txt decide if useragent can fetch url"""
        if self.disallow_all:
            return 0
        if self.allow_all:
            return 1
        # search for given user agent matches
        # the first match counts
        url = urllib.quote(urlparse.urlparse(urllib.unquote(url))[2]) or "/"
        for entry in self.entries:
            if entry.applies_to(useragent):
                if not entry.allowance(url):
                    return 0
        # agent not found ==> access granted
        return 1
    
class URLGatekeeper:
    """a class to track robots.txt rules across multiple servers"""
    def __init__(self):
        self.rpcache = {} # a dictionary of RobotFileParserFixed objects, by domain
        self.urlopener = urllib.FancyURLopener()
        self.urlopener.version = "feedfinder/" + __version__ + " " + self.urlopener.version + " +http://diveintomark.org/projects/feed_finder/"
        _debuglog(self.urlopener.version)
        self.urlopener.addheaders = [('User-agent', self.urlopener.version)]
        robotparser.URLopener.version = self.urlopener.version
        robotparser.URLopener.addheaders = self.urlopener.addheaders
        
    def _getrp(self, url):
        protocol, domain = urlparse.urlparse(url)[:2]
        if self.rpcache.has_key(domain):
            return self.rpcache[domain]
        baseurl = '%s://%s' % (protocol, domain)
        robotsurl = urlparse.urljoin(baseurl, 'robots.txt')
        _debuglog('fetching %s' % robotsurl)
        rp = RobotFileParserFixed(robotsurl)
        rp.read()
        self.rpcache[domain] = rp
        return rp
        
    def can_fetch(self, url):
        rp = self._getrp(url)
        allow = rp.can_fetch(self.urlopener.version, url)
        _debuglog("Gatekeeper examined %s and said %s" % (url, allow))
        return allow

    def get(self, url):
        if not self.can_fetch(url): return ''
        return self.urlopener.open(url).read()

_gatekeeper = URLGatekeeper()

class BaseParser(sgmllib.SGMLParser):
    def __init__(self, baseuri):
        sgmllib.SGMLParser.__init__(self)
        self.links = []
        self.baseuri = baseuri
        
    def normalize_attrs(self, attrs):
        attrs = [(k.lower(), sgmllib.charref.sub(lambda m: chr(int(m.groups()[0])), v).strip()) for k, v in attrs]
        attrs = [(k, k in ('rel','type') and v.lower() or v) for k, v in attrs]
        return attrs
        
    def do_base(self, attrs):
        attrsD = dict(self.normalize_attrs(attrs))
        if not attrsD.has_key('href'): return
        self.baseuri = attrsD['href']
        
class LinkParser(BaseParser):
    FEED_TYPES = ('application/rss+xml',
                  'text/xml',
                  'application/atom+xml',
                  'application/x.atom+xml',
                  'application/x-atom+xml')
    def do_link(self, attrs):
        attrsD = dict(self.normalize_attrs(attrs))
        if not attrsD.has_key('rel'): return
        rels = attrsD['rel'].split()
        if 'alternate' not in rels: return
        if attrsD.get('type') not in self.FEED_TYPES: return
        if not attrsD.has_key('href'): return
        self.links.append(urlparse.urljoin(self.baseuri, attrsD['href']))

class ALinkParser(BaseParser):
    def start_a(self, attrs):
        attrsD = dict(self.normalize_attrs(attrs))
        if not attrsD.has_key('href'): return
        self.links.append(urlparse.urljoin(self.baseuri, attrsD['href']))

def makeFullURI(uri):
    if (not uri.startswith('http://')) and (not uri.startswith('https://')):
        uri = 'http://%s' % uri
    return uri

def getLinks(data, baseuri):
    p = LinkParser(baseuri)
    p.feed(data)
    return p.links

def getALinks(data, baseuri):
    p = ALinkParser(baseuri)
    p.feed(data)
    return p.links

def getLocalLinks(links, baseuri):
    baseuri = baseuri.lower()
    urilen = len(baseuri)
    return [l for l in links if l.lower().startswith(baseuri)]

def isFeedLink(link):
    return link[-4:].lower() in ('.rss', '.rdf', '.xml', '.atom')

def isXMLRelatedLink(link):
    link = link.lower()
    return link.count('rss') + link.count('rdf') + link.count('xml') + link.count('atom')

def couldBeFeedData(data):
    data = data.lower()
    #if data.count('<html'):
    count = 0
    count = data.count('<rss') + data.count('<rdf') + data.count('<feed')
    return count

def isFeed(uri):
    _debuglog('verifying that %s is a feed' % uri)
    protocol = urlparse.urlparse(uri)
    if protocol[0] not in ('http', 'https'): return 0
    data = _gatekeeper.get(uri)
    return couldBeFeedData(data)

def sortFeeds(feed1Info, feed2Info):
    return cmp(feed2Info['headlines_rank'], feed1Info['headlines_rank'])

def getFeedsFromSyndic8(uri):
    feeds = []
    try:
        server = xmlrpclib.Server('http://www.syndic8.com/xmlrpc.php')
        feedids = server.syndic8.FindFeeds(uri)
        infolist = server.syndic8.GetFeedInfo(feedids, ['headlines_rank','status','dataurl'])
        infolist.sort(sortFeeds)
        feeds = [f['dataurl'] for f in infolist if f['status']=='Syndicated']
        _debuglog('found %s feeds through Syndic8' % len(feeds))
    except:
        pass
    return feeds
    
def getFeeds(uri, querySyndic8=0):
    fulluri = makeFullURI(uri)
    data = _gatekeeper.get(fulluri)
    # is this already a feed?
    if couldBeFeedData(data):
        return [fulluri]
    # nope, it's a page, try LINK tags first
    _debuglog('looking for LINK tags')
    feeds = getLinks(data, fulluri)
    _debuglog('found %s feeds through LINK tags' % len(feeds))
    feeds = filter(isFeed, feeds)
    if not feeds:
        # no LINK tags, look for regular <A> links that point to feeds
        _debuglog('no LINK tags, looking at A tags')
        links = getALinks(data, fulluri)
        locallinks = getLocalLinks(links, fulluri)
        # look for obvious feed links on the same server
        feeds = filter(isFeed, filter(isFeedLink, locallinks))
        if not feeds:
            # look harder for feed links on the same server
            feeds = filter(isFeed, filter(isXMLRelatedLink, locallinks))
        if not feeds:
            # look for obvious feed links on another server
            feeds = filter(isFeed, filter(isFeedLink, links))
        if not feeds:
            # look harder for feed links on another server
            feeds = filter(isFeed, filter(isXMLRelatedLink, links))
    if not feeds and querySyndic8:
        # still no luck, search Syndic8 for feeds (requires xmlrpclib)
        _debuglog('still no luck, searching Syndic8')
        feeds = getFeedsFromSyndic8(uri)
    return feeds

##### test harness ######

def test():
    uri = 'http://diveintomark.org/tests/client/autodiscovery/html4-001.html'
    failed = []
    count = 0
    while 1:
        data = urllib.urlopen(uri).read()
        if data.find('Atom autodiscovery test') == -1: break
        sys.stdout.write('.')
        count += 1
        links = getLinks(data, uri)
        if not links:
            print '\n*** FAILED ***', uri, 'could not find link'
            failed.append(uri)
        elif len(links) > 1:
            print '\n*** FAILED ***', uri, 'found too many links'
            failed.append(uri)
        else:
            atomdata = urllib.urlopen(links[0]).read()
            if atomdata.find('<link rel="alternate"') == -1:
                print '\n*** FAILED ***', uri, 'retrieved something that is not a feed'
                failed.append(uri)
            else:
                backlink = atomdata.split('href="').pop().split('"')[0]
                if backlink != uri:
                    print '\n*** FAILED ***', uri, 'retrieved wrong feed'
                    failed.append(uri)
        if data.find('<link rel="next" href="') == -1: break
        uri = urlparse.urljoin(uri, data.split('<link rel="next" href="').pop().split('"')[0])
    print
    print count, 'tests executed,', len(failed), 'failed'
        
if __name__ == '__main__':
    if sys.argv[1:]:
        uri = sys.argv[1]
    else:
        uri = 'http://diveintomark.org/'
    if uri == 'test':
        test()
    else:
        print "\n".join(getFeeds(uri))
